from math import sqrt, sin, cos, pi, asin, pow, ceil, log
import numpy as np
from scipy import stats
import pandas as pd
from collections import defaultdict
import sys
from tqdm import tqdm
from skmob.utils.gislib import getDistanceByHaversine
tqdm.pandas()
from ..utils import constants

def _radius_of_gyration_individual(traj):
    """
    Compute the radius of gyration of a single individual given their TrajDataFrame.

    Parameters
    ----------
    traj : TrajDataFrame
        the trajectory of the individual.
    
    Returns
    -------
    float
        the radius of gyration of the individual.
    """
    lats_lngs = traj[[constants.LATITUDE, constants.LONGITUDE]].values
    center_of_mass = np.mean(lats_lngs, axis=0)
    rg = np.sqrt(np.mean([getDistanceByHaversine((lat, lng), center_of_mass) ** 2.0 for lat, lng in lats_lngs]))
    return rg


def radius_of_gyration(traj, show_progress=True):
    """Radius of gyration.
    
    Compute the radii of gyration (in kilometers) of a set of individuals in a TrajDataFrame.
    The radius of gyration of an individual :math:`u` is defined as [GHB2008]_ [PRQPG2013]_: 
    
    .. math:: 
        r_g(u) = \sqrt{ \\frac{1}{n_u} \sum_{i=1}^{n_u} dist(r_i(u) - r_{cm}(u))^2}
    
    where :math:`r_i(u)` represents the :math:`n_u` positions recorded for :math:`u`, and :math:`r_{cm}(u)` is the center of mass of :math:`u`'s trajectory. In mobility analysis, the radius of gyration indicates the characteristic distance travelled by :math:`u`.

    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individuals.
    
    show_progress : boolean, optional
        if True, show a progress bar. The default is True.
    
    Returns
    -------
    pandas DataFrame
        the radius of gyration of each individual.
    
    Examples
    --------
    >>> import skmob
    >>> from skmob.measures.individual import radius_of_gyration
    >>> url = "https://snap.stanford.edu/data/loc-brightkite_totalCheckins.txt.gz"
    >>> df = pd.read_csv(url, sep='\\t', header=0, nrows=100000, 
                 names=['user', 'check-in_time', 'latitude', 'longitude', 'location id'])
    >>> tdf = skmob.TrajDataFrame(df, latitude='latitude', longitude='longitude', datetime='check-in_time', user_id='user')
    >>> rg_df = radius_of_gyration(tdf)
    >>> print(rg_df.head())
       uid  radius_of_gyration
    0    0         1564.436792
    1    1         2467.773523
    2    2         1439.649774
    3    3         1752.604191
    4    4         5380.503250    

    References
    ----------
    .. [GHB2008] González, M. C., Hidalgo, C. A. & Barabási, A. L. (2008) Understanding individual human mobility patterns. Nature, 453, 779–782, https://www.nature.com/articles/nature06958.
    .. [PRQPG2013] Pappalardo, L., Rinzivillo, S., Qu, Z., Pedreschi, D. & Giannotti, F. (2013) Understanding the patterns of car travel. European Physics Journal Special Topics 215(1), 61-73, https://link.springer.com/article/10.1140%2Fepjst%2Fe2013-01715-5

    See Also
    --------
    k_radius_of_gyration
    """
    # if 'uid' column in not present in the TrajDataFrame
    if constants.UID not in traj.columns:
        return pd.DataFrame([_radius_of_gyration_individual(traj)], columns=[sys._getframe().f_code.co_name])
    
    if show_progress:
        df = traj.groupby(constants.UID).progress_apply(lambda x: _radius_of_gyration_individual(x))
    else:
        df = traj.groupby(constants.UID).apply(lambda x: _radius_of_gyration_individual(x))
    return pd.DataFrame(df).reset_index().rename(columns={0: sys._getframe().f_code.co_name}) 


def _k_radius_of_gyration_individual(traj, k=2):
    """Compute the k-radius of gyration of a single individual given their TrajDataFrame.

    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individual.
    
    k : int, optional
        the number of most frequent locations to consider. The default is 2. The possible range of values is math:`[2, +inf]`.
    
    Returns
    -------
    float
        the k-radius of gyration of the individual. 
    """
    traj['visits'] = traj.groupby([constants.LATITUDE, constants.LONGITUDE]).transform('count')[constants.DATETIME]
    top_k_locations = traj.drop_duplicates(subset=[constants.LATITUDE, constants.LONGITUDE]).sort_values(by=['visits', constants.DATETIME],
                                                                              ascending=[False, True])[:k]
    visits = top_k_locations['visits'].values
    total_visits = sum(visits)
    lats_lngs = top_k_locations[[constants.LATITUDE, constants.LONGITUDE]].values

    center_of_mass = visits.dot(lats_lngs) / total_visits
    krg = np.sqrt(sum([visits[i] * (getDistanceByHaversine((lat, lng), center_of_mass) ** 2.0)
                       for i, (lat, lng) in enumerate(lats_lngs)]) / total_visits)
    return krg


def k_radius_of_gyration(traj, k=2, show_progress=True):
    """k-radius of gyration.
    
    Compute the k-radii of gyration (in kilometers) of a set of individuals in a TrajDataFrame.
    The k-radius of gyration of an individual :math:`u` is defined as [PSRPGB2015]_:
    
    .. math::
        r_g^{(k)}(u) = \sqrt{\\frac{1}{n_u^{(k)}} \sum_{i=1}^k (r_i(u) - r_{cm}^{(k)}(u))^2} 
        
    where :math:`r_i(u)` represents the :math:`n_u^{(k)}` positions recorded for :math:`u` on their k most frequent locations, and :math:`r_{cm}^{(k)}(u)` is the center of mass of :math:`u`'s trajectory considering the visits to the k most frequent locations only. In mobility analysis, the k-radius of gyration indicates the characteristic distance travelled by that individual as induced by their k most frequent locations.

    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individual.
    
    k : int, optional
        the number of most frequent locations to consider. The default is 2. The possible range of values is :math:`[2, +inf]`.
    
    show_progress : boolean, optional
        if True, show a progress bar. The default is True.
    
    Returns
    -------
    pandas DataFrame
        the k-radii of gyration of the individuals

    Examples
    --------
    >>> import skmob
    >>> from skmob.measures.individual import k_radius_of_gyration
    >>> url = "https://snap.stanford.edu/data/loc-brightkite_totalCheckins.txt.gz"
    >>> df = pd.read_csv(url, sep='\\t', header=0, nrows=100000, 
                 names=['user', 'check-in_time', 'latitude', 'longitude', 'location id'])
    >>> tdf = skmob.TrajDataFrame(df, latitude='latitude', longitude='longitude', datetime='check-in_time', user_id='user')
    >>> krg_df = k_radius_of_gyration(tdf)
    >>> print(krg_df.head())
       uid  3k_radius_of_gyration
    0    0               7.730516
    1    1               3.620671
    2    2               6.366549
    3    3              10.543072
    4    4            3910.808802

    References
    ----------
    .. [PSRPGB2015] Pappalardo, L., Simini, F. Rinzivillo, S., Pedreschi, D. Giannotti, F. & Barabasi, A. L. (2015) Returners and Explorers dichotomy in human mobility. Nature Communications 6, https://www.nature.com/articles/ncomms9166
    
    See Also
    --------
    radius_of_gyration
    """
    # if 'uid' column in not present in the TrajDataFrame
    if constants.UID not in traj.columns:
        return pd.DataFrame([_k_radius_of_gyration_individual(traj, k=k)], columns=['%s%s' % (k, sys._getframe().f_code.co_name)])
    
    if show_progress:
        df = traj.groupby(constants.UID).progress_apply(lambda x: _k_radius_of_gyration_individual(x, k))
    else:
        df = traj.groupby(constants.UID).apply(lambda x: _k_radius_of_gyration_individual(x, k))
    return pd.DataFrame(df).reset_index().rename(columns={0: '%s%s' % (k, sys._getframe().f_code.co_name)})


def _random_entropy_individual(traj):
    """
    Compute the random entropy of a single individual given their TrajDataFrame.
    
    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individual
    
    Returns
    -------
    float
        the random entropy of the individual 
    """
    n_distinct_locs = len(traj.groupby([constants.LATITUDE, constants.LONGITUDE]))
    entropy = np.log2(n_distinct_locs)
    return entropy


def random_entropy(traj, show_progress=True):
    """Random entropy.
    
    Compute the random entropy of a set of individuals in a TrajDataFrame.
    The random entropy of an individual :math:`u` is defined as [EP2009]_ [SQBB2010]_: 
    
    .. math::
        E_{rand}(u) = log_2(N_u)
    
    where :math:`N_u` is the number of distinct locations visited by :math:`u`, capturing the degree of predictability of :math:`u`’s whereabouts if each location is visited with equal probability. 

    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individuals.
    
    show_progress : boolean, optional
        if True, show a progress bar. The default is True.
    
    Returns
    -------
    pandas DataFrame
        the random entropy of the individuals.
    
    Examples
    --------
    >>> import skmob
    >>> from skmob.measures.individual import random_entropy
    >>> url = "https://snap.stanford.edu/data/loc-brightkite_totalCheckins.txt.gz"
    >>> df = pd.read_csv(url, sep='\\t', header=0, nrows=100000, 
                 names=['user', 'check-in_time', 'latitude', 'longitude', 'location id'])
    >>> tdf = skmob.TrajDataFrame(df, latitude='latitude', longitude='longitude', datetime='check-in_time', user_id='user')
    >>> re_df = random_entropy(tdf)
    >>> print(re_df.head())
       uid  random_entropy
    0    0        9.082149
    1    1        6.599913
    2    2        8.845490
    3    3        9.262095
    4    4        7.754888

    References
    ----------
    .. [EP2009] Eagle, N. & Pentland, A. S. (2009) Eigenbehaviors: identifying structure in routine. Behavioral Ecology and Sociobiology 63(7), 1057-1066, https://link.springer.com/article/10.1007/s00265-009-0830-6
    .. [SQBB2010] Song, C., Qu, Z., Blumm, N. & Barabási, A. L. (2010) Limits of Predictability in Human Mobility. Science 327(5968), 1018-1021, https://science.sciencemag.org/content/327/5968/1018
    
    See Also
    --------
    uncorrelated_entropy, real_entropy
    """
    # if 'uid' column in not present in the TrajDataFrame
    if constants.UID not in traj.columns:
        return pd.DataFrame([_random_entropy_individual(traj)], columns=[sys._getframe().f_code.co_name])
    
    if show_progress:
        df = traj.groupby(constants.UID).progress_apply(lambda x: _random_entropy_individual(x))
    else:
        df = traj.groupby(constants.UID).apply(lambda x: _random_entropy_individual(x))
    return pd.DataFrame(df).reset_index().rename(columns={0: sys._getframe().f_code.co_name})


def _uncorrelated_entropy_individual(traj, normalize=False):
    """
    Compute the uncorrelated entropy of a single individual given their TrajDataFrame.

    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individuals.
    
    normalize : boolean, optional
        if True, normalize the entropy in the range :math:`[0, 1]` by dividing by :math:`log_2(N_u)`, where :math:`N` is the number of distinct locations visited by individual :math:`u`. The default is False.

    Returns
    -------
    float
        the temporal-uncorrelated entropy of the individual
    """
    n = len(traj)
    probs = [1.0 * len(group) / n for group in traj.groupby(by=[constants.LATITUDE, constants.LONGITUDE]).groups.values()]
    entropy = stats.entropy(probs, base=2.0)
    if normalize:
        n_vals = len(np.unique(traj[[constants.LATITUDE, constants.LONGITUDE]].values, axis=0))
        if n_vals > 1:
            entropy /= np.log2(n_vals)
        else:  # to avoid NaN
            entropy = 0.0
    return entropy


def uncorrelated_entropy(traj, normalize=False, show_progress=True):
    """Uncorrelated entropy.
    
    Compute the temporal-uncorrelated entropy of a set of individuals in a TrajDataFrame. The temporal-uncorrelated entropy of an individual :math:`u` is defined as [EP2009]_ [SQBB2010]_ [PVGSPG2016]_: 
    
    .. math::
        E_{unc}(u) = - \sum_{j=1}^{N_u} p_u(j) log_2 p_u(j)
    
    where :math:`N_u` is the number of distinct locations visited by :math:`u` and :math:`p_u(j)` is the historical probability that a location :math:`j` was visited by :math:`u`. The temporal-uncorrelated entropy characterizes the heterogeneity of :math:`u`'s visitation patterns.
    
    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individuals.
    
    normalize : boolean, optional
        if True, normalize the entropy in the range :math:`[0, 1]` by dividing by :math:`log_2(N_u)`, where :math:`N` is the number of distinct locations visited by individual :math:`u`. The default is False.
    
    show_progress : boolean, optional
        if True, show a progress bar. The default is True.
    
    Returns
    -------
    pandas DataFrame
        the temporal-uncorrelated entropy of the individuals.
    
    Examples
    --------
    >>> import skmob
    >>> from skmob.measures.individual import uncorrelated_entropy
    >>> url = "https://snap.stanford.edu/data/loc-brightkite_totalCheckins.txt.gz"
    >>> df = pd.read_csv(url, sep='\\t', header=0, nrows=100000, 
                 names=['user', 'check-in_time', 'latitude', 'longitude', 'location id'])
    >>> tdf = skmob.TrajDataFrame(df, latitude='latitude', longitude='longitude', datetime='check-in_time', user_id='user')
    >>> ue_df = uncorrelated_entropy(tdf, normalize=True)
    >>> print(ue_df.head())
       uid  norm_uncorrelated_entropy
    0    0                   0.819430
    1    1                   0.552972
    2    2                   0.764304
    3    3                   0.794553
    4    4                   0.756421

    References
    ----------
    .. [PVGSPG2016] Pappalardo, L., Vanhoof, M., Gabrielli, L., Smoreda, Z., Pedreschi, D. & Giannotti, F. (2016) An analytical framework to nowcast well-being using mobile phone data. International Journal of Data Science and Analytics 2(75), 75-92, https://link.springer.com/article/10.1007/s41060-016-0013-2
    
    See Also
    --------
    random_entropy, real_entropy
    """
    column_name = sys._getframe().f_code.co_name
    if normalize:
        column_name = 'norm_%s' % sys._getframe().f_code.co_name
    
    # if 'uid' column in not present in the TrajDataFrame
    if constants.UID not in traj.columns:
        return pd.DataFrame([_uncorrelated_entropy_individual(traj)], columns=[column_name])
    
    if show_progress:
        df = traj.groupby(constants.UID).progress_apply(lambda x: _uncorrelated_entropy_individual(x, normalize=normalize))
    else:
        df = traj.groupby(constants.UID).apply(lambda x: _uncorrelated_entropy_individual(x, normalize=normalize))
    return pd.DataFrame(df).reset_index().rename(columns={0: column_name})


def _stringify(seq):
    return '|'.join(['_'.join(list(map(str, r))) for r in seq])


def _true_entropy(sequence):
    n = len(sequence)

    # these are the first and last elements
    sum_lambda = 1. + 2.

    for i in range(1, n - 1):
        str_seq = _stringify(sequence[:i])
        j = 1
        str_sub_seq = _stringify(sequence[i:i + j])
        while str_sub_seq in str_seq:
            j += 1
            str_sub_seq = _stringify(sequence[i:i + j])
            if i + j == n:
                # EOF character
                j += 1
                break
        sum_lambda += j

    return 1. / sum_lambda * n * np.log2(n)


def _real_entropy_individual(traj):
    """
    Compute the real entropy of a single individual given their TrajDataFrame.

    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individual.
    
    Returns
    -------
    float
        the real entropy of the individual.
    """
    time_series = tuple(map(tuple, traj[[constants.LATITUDE, constants.LONGITUDE]].values))
    entropy = _true_entropy(time_series)
    return entropy


def real_entropy(traj, show_progress=True):
    """Real entropy.
    
    Compute the real entropy of a set of individuals in a TrajDataFrame. 
    The real entropy of an individual :math:`u` is defined as [SQBB2010]_: 
    
    .. math:: 
        E(u) = - \sum_{T'_u}P(T'_u)log_2[P(T_u^i)]
    
    where :math:`P(T'_u)` is the probability of finding a particular time-ordered subsequence :math:`T'_u` in the trajectory :math:`T_u`. The real entropy hence depends not only on the frequency of visitation, but also the order in which the nodes were visited and the time spent at each location, thus capturing the full spatiotemporal order present in an :math:`u`'s mobility patterns.
    
    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individuals
    
    show_progress : boolean, optional 
        if True, show a progress bar. The default is True.
    
    Returns
    -------
    pandas DataFrame
        the real entropy of the individuals
    
    Warning
    -------
    The input TrajDataFrame must be sorted in ascending order by `datetime`. Note that the computation of this measure is, by construction, slow.
    
    Examples
    --------
    >>> import skmob
    >>> from skmob.measures.individual import real_entropy
    >>> url = "https://snap.stanford.edu/data/loc-brightkite_totalCheckins.txt.gz"
    >>> df = pd.read_csv(url, sep='\\t', header=0, nrows=100000, 
                 names=['user', 'check-in_time', 'latitude', 'longitude', 'location id'])
    >>> tdf = skmob.TrajDataFrame(df, latitude='latitude', longitude='longitude', datetime='check-in_time', user_id='user')
    >>> re_df = real_entropy(tdf[tdf.uid < 50]) # computed on a subset of individuals
    >>> print(re_df.head())
       uid  real_entropy
    0    0      4.906479
    1    1      2.207224
    2    2      4.467225
    3    3      4.782442
    4    4      3.585371

    See Also
    --------
    random_entropy, uncorrelated_entropy
    """
    # if 'uid' column in not present in the TrajDataFrame
    if constants.UID not in traj.columns:
        return pd.DataFrame([_real_entropy_individual(traj)], columns=[sys._getframe().f_code.co_name])
    
    if show_progress:
        df = traj.groupby(constants.UID).progress_apply(lambda x: _real_entropy_individual(x))
    else:
        df = traj.groupby(constants.UID).apply(lambda x: _real_entropy_individual(x))
    return pd.DataFrame(df).reset_index().rename(columns={0: sys._getframe().f_code.co_name})


def _jump_lengths_individual(traj):
    """
    Compute the jump lengths (in kilometers) of a single individual from their TrajDataFrame.
    
    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individual.
    
    Returns
    -------
    list
        the list of distances (in kilometers) traveled by the individual.
    """
    if len(traj) == 1:  # if there is just one point, no distance can be computed
        return []
    lats_lngs = traj.sort_values(by=constants.DATETIME)[[constants.LATITUDE, constants.LONGITUDE]].values
    lengths = np.array([getDistanceByHaversine(lats_lngs[i], lats_lngs[i - 1]) for i in range(1, len(lats_lngs))])
    return lengths


def jump_lengths(traj, show_progress=True, merge=False):
    """Jump lengths.
    
    Compute the jump lengths (in kilometers) of a set of individuals in a TrajDataFrame.
    A jump length (or trip distance) :math:`\Delta r`made by an individual :math:`u` is defined as the geographic distance between two consecutive points visited by :math:`u`: 
    
    .. math:: \Delta r = dist(r_i, r_{i + 1})
    
    where :math:`r_i` and :math:`r_{i + 1}` are two consecutive points, described as a latitude, longitude pair, in the time-ordered trajectory of an individual, and :math:`dist` is the geographic distance between the two points [BHG2006]_ [GHB2008]_ [PRQPG2013]_.

    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individuals
    
    show_progress : boolean, optional
        if True, show a progress bar. The default is True.
    
    merge : boolean, optional
        if True, merge the individuals' lists into one list. The default is False.
    
    Returns
    -------
    pandas DataFrame or list
        the jump lengths for each individual, where :math:`NaN` indicates that an individual visited just one location and hence distance is not defined; or a list with all jumps together if `merge` is True.

    Warning
    -------
    The input TrajDataFrame must be sorted in ascending order by `datetime`. 

    Examples
    --------
    >>> import skmob
    >>> from skmob.measures.individual import jump_lengths
    >>> url = "https://snap.stanford.edu/data/loc-brightkite_totalCheckins.txt.gz"
    >>> df = pd.read_csv(url, sep='\\t', header=0, nrows=100000, 
                 names=['user', 'check-in_time', 'latitude', 'longitude', 'location id'])
    >>> tdf = skmob.TrajDataFrame(df, latitude='latitude', longitude='longitude', datetime='check-in_time', user_id='user')
    >>> jl_df = jump_lengths(tdf)
    >>> print(jl_df.head())
       uid                                       jump_lengths
    0    0  [19.640467328877936, 0.0, 0.0, 1.7434311010381...
    1    1  [6.505330424378251, 46.75436600375988, 53.9284...
    2    2  [0.0, 0.0, 0.0, 0.0, 3.6410097195943507, 0.0, ...
    3    3  [3861.2706300798827, 4.061631313492122, 5.9163...
    4    4  [15511.92758595804, 0.0, 15511.92758595804, 1....
    >>> jl_list = jump_lengths(tdf, merge=True)
    >>> print(jl_list[:10]) # print the first ten elements in the list
    [19.640467328877936, 0.0, 0.0, 1.743431101038163, 1553.5011134765616, 0.0, 30.14517724008101, 0.0, 2.563647571198179, 1.9309489380903868]
    
    References
    ----------
    .. [BHG2006] Brockmann, D., Hufnagel, L. & Geisel, T. (2006) The scaling laws of human travel. Nature 439, 462-465, https://www.nature.com/articles/nature04292
    
    See Also
    --------
    maximum_distance, distance_straight_line
    """
    # if 'uid' column in not present in the TrajDataFrame
    if constants.UID not in traj.columns:
        return pd.DataFrame(pd.Series([_jump_lengths_individual(traj)]), columns=[sys._getframe().f_code.co_name])
    
    if show_progress:
        df = traj.groupby(constants.UID).progress_apply(lambda x: _jump_lengths_individual(x))
    else:
        df = traj.groupby(constants.UID).apply(lambda x: _jump_lengths_individual(x))
    
    df = pd.DataFrame(df).reset_index().rename(columns={0: sys._getframe().f_code.co_name})
    
    if merge:
        # merge all lists 
        jl_list =[]
        for x in df.jump_lengths:
            jl_list.extend(x)
        return jl_list
    return df


def _maximum_distance_individual(traj):
    """
    Compute the maximum distance (in kilometers) traveled by an individual given their TrajDataFrame.
    
    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individual
    
    Returns
    -------
    float
        the maximum traveled distance for the individual. Note that :math:`NaN` indicates that an individual visited just one location and hence distance is not defined.
    """
    jumps = _jump_lengths_individual(traj)
    if len(jumps) > 0:
        return max(jumps)
    return np.NaN

def maximum_distance(traj, show_progress=True):
    """Maximum distance.
    
    Compute the maximum distance (in kilometers) traveled by a set of individuals in a TrajDataFrame. The maximum distance :math:`d_{max}` travelled by an individual :math:`u` is defined as: 
    
    .. math:: d_{max} = \max\limits_{1 \leq i \lt j \lt n_u} dist(r_i, r_j)
    
    where :math:`n_u` is the number of points recorded for :math:`u`, :math:`r_i` and :math:`r_{i + 1}` are two consecutive points, described as a :math:`(latitude, longitude)` pair, in :math:`u`'s time-ordered trajectory, and :math:`dist` is the geographic distance between the two points [WTDED2015]_ [LBH2012]_.

    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individuals.
    
    show_progress : boolean, optional
        if True, show a progress bar. The default is True.
    
    Returns
    -------
    pandas DataFrame
        the maximum traveled distance for each individual. Note that :math:`NaN` indicates that an individual visited just one location and so the maximum distance is not defined.

    Examples
    --------
    >>> import skmob
    >>> from skmob.measures.individual import maximum_distance
    >>> url = "https://snap.stanford.edu/data/loc-brightkite_totalCheckins.txt.gz"
    >>> df = pd.read_csv(url, sep='\\t', header=0, nrows=100000, 
                 names=['user', 'check-in_time', 'latitude', 'longitude', 'location id'])
    >>> tdf = skmob.TrajDataFrame(df, latitude='latitude', longitude='longitude', datetime='check-in_time', user_id='user')
    >>> md_df = maximum_distance(tdf)
    >>> print(md_df.head())
       uid  maximum_distance
    0    0      11294.436420
    1    1      12804.895064
    2    2      11286.745660
    3    3      12803.259219
    4    4      15511.927586

    References
    ----------
    .. [WTDED2015] Williams, N. E., Thomas, T. A., Dunbar, M., Eagle, N. & Dobra, A. (2015) Measures of Human Mobility Using Mobile Phone Records Enhanced with GIS Data. PLOS ONE 10(7): e0133630. https://doi.org/10.1371/journal.pone.0133630
    .. [LBH2012] Lu, X., Bengtsson, L. & Holme, P. (2012) Predictability of population displacement after the 2010 haiti earthquake. Proceedings of the National Academy of Sciences 109 (29) 11576-11581; https://doi.org/10.1073/pnas.1203882109
    
    See Also
    --------
    jump_lengths, distance_straight_line
    """
    # if 'uid' column in not present in the TrajDataFrame
    if constants.UID not in traj.columns:
        return pd.DataFrame([_maximum_distance_individual(traj)], columns=[sys._getframe().f_code.co_name])
    
    if show_progress:
        df = traj.groupby(constants.UID).progress_apply(lambda x: _maximum_distance_individual(x))
    else:
        df = traj.groupby(constants.UID).apply(lambda x: _maximum_distance_individual(x))
    return pd.DataFrame(df).reset_index().rename(columns={0: sys._getframe().f_code.co_name})

def _distance_straight_line_individual(traj):
    """
    Compute the distance straight line travelled by the individual given their TrajDataFrame.
    
    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individual
    
    Returns
    -------
    float
        the straight line distance traveled by the individual. Note the :math:`NaN` indicates that the individual visited just one location and hence distance is not defined.
    """
    jumps = _jump_lengths_individual(traj)
    if len(jumps) > 0:
        return sum(jumps)
    return 0.0

def distance_straight_line(traj, show_progress=True):
    """Distance straight line.
    
    Compute the distance (in kilometers) travelled straight line by a set of individuals in a TrajDataFrame. The distance straight line :math:`d_{SL}` travelled by an individual :math:`u` is computed as the sum of the distances travelled :math:`u`: 
    
    .. math:: d_{SL} = \sum_{j=2}^{n_u} dist(r_{j-1}, r_j)
    
    where :math:`n_u` is the number of points recorded for :math:`u`, :math:`r_{j-1}` and :math:`r_j` are two consecutive points, described as a :math:`(latitude, longitude)` pair, in :math:`u`'s time-ordered trajectory, and :math:`dist` is the geographic distance between the two points [WTDED2015]_.
    
    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individuals.
    
    show_progress : boolean, optional 
        if True, show a progress bar. The default is True.
    
    Returns
    -------
    pandas DataFrame
        the straight line distance traveled by the individuals. Note that :math:`NaN` indicates that an individual visited just one location and hence distance is not defined.

    Warning
    -------
    The input TrajDataFrame must be sorted in ascending order by `datetime`.

    Examples
    --------
    >>> import skmob
    >>> from skmob.measures.individual import distance_straight_line
    >>> url = "https://snap.stanford.edu/data/loc-brightkite_totalCheckins.txt.gz"
    >>> df = pd.read_csv(url, sep='\\t', header=0, nrows=100000, 
                 names=['user', 'check-in_time', 'latitude', 'longitude', 'location id'])
    >>> tdf = skmob.TrajDataFrame(df, latitude='latitude', longitude='longitude', datetime='check-in_time', user_id='user')
    >>> dsl_df = distance_straight_line(tdf)
    >>> print(dsl_df.head())
       uid  distance_straight_line
    0    0           374530.954882
    1    1           774346.816009
    2    2            88710.682464
    3    3           470986.771764
    4    4           214623.524252

    See Also
    --------
    jump_lengths, maximum_distance
    """
    # if 'uid' column in not present in the TrajDataFrame
    if constants.UID not in traj.columns:
        return pd.DataFrame([_distance_straight_line_individual(traj)], columns=[sys._getframe().f_code.co_name])
    
    if show_progress:
        df = traj.groupby(constants.UID).progress_apply(lambda x: _distance_straight_line_individual(x))
    else:
        df = traj.groupby(constants.UID).apply(lambda x: _distance_straight_line_individual(x))
    return pd.DataFrame(df).reset_index().rename(columns={0: sys._getframe().f_code.co_name})


def _waiting_times_individual(traj):
    """
    Compute the waiting times for a single individual given their TrajDataFrame.
    
    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individual.
    
    Returns
    -------
    list
        the waiting times of the individual.
    """
    if len(traj) == 1:
        return []
    times = traj.sort_values(by=constants.DATETIME)[constants.DATETIME]
    wtimes = times.diff()[1:].values.astype('timedelta64[s]').astype('float')
    return wtimes


def waiting_times(traj, show_progress=True, merge=False):
    """Waiting times.
    
    Compute the waiting times (in seconds) between the movements of each individual in a TrajDataFrame. A waiting time (or inter-time) by an individual :math:`u` is defined as the time between two consecutive points in :math:`u`'s trajectory:
    
    .. math:: \Delta t = |t(r_i) - t(r_{i + 1})|
    
    where :math:`r_i` and :math:`r_{i + 1}` are two consecutive points, described as a :math:`(latitude, longitude)` pair, in the time-ordered trajectory of :math:`u`, and :math:`t(r)` indicates the time when :math:`u` visits point :math:`r` [SKWB2010]_ [PF2018]_.
    
    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individuals.
    
    show_progress : boolean, optional
        if True, show a progress bar. The default is True.
    
    merge : boolean, optional
        if True, merge the individuals' lists into one list. The default is False.
    
    Returns
    -------
    pandas DataFrame or list
        the list of waiting times for each individual, where :math:`NaN` indicates that an individual visited just one location and hence waiting time is not defined; or a list with all waiting times together if `merge` is True.
    
    Warning
    -------
    The input TrajDataFrame must by sorted in ascending order by `datetime`.
    
    Examples
    --------
    >>> import skmob
    >>> from skmob.measures.individual import waiting_times
    >>> url = "https://snap.stanford.edu/data/loc-brightkite_totalCheckins.txt.gz"
    >>> df = pd.read_csv(url, sep='\\t', header=0, nrows=100000, 
                 names=['user', 'check-in_time', 'latitude', 'longitude', 'location id'])
    >>> tdf = skmob.TrajDataFrame(df, latitude='latitude', longitude='longitude', datetime='check-in_time', user_id='user')
    >>> wt_df = waiting_times(tdf)
    >>> print(wt_df.head())
       uid                                      waiting_times
    0    0  [2358.0, 136.0, 303.0, 1836.0, 14869.0, 517.0,...
    1    1  [43460.0, 34353.0, 8347.0, 40694.0, 281.0, 16....
    2    2  [293.0, 308.0, 228.0, 402.0, 16086.0, 665.0, 9...
    3    3  [10200079.0, 30864.0, 54415.0, 2135.0, 63.0, 1...
    4    4  [82845.0, 56.0, 415156.0, 1372.0, 23.0, 42679....
    >>> wl_list = waiting_times(tdf, merge=True)
    >>> print(wl_list[:10])
    [2358.0, 136.0, 303.0, 1836.0, 14869.0, 517.0, 8995.0, 41306.0, 949.0, 11782.0]
    
    References
    ----------
    .. [SKWB2010] Song, C., Koren, T., Wang, P. & Barabasi, A.L. (2010) Modelling the scaling properties of human mobility. Nature Physics 6, 818-823, https://www.nature.com/articles/nphys1760
    .. [PF2018] Pappalardo, L. & Simini, F. (2018) Data-driven generation of spatio-temporal routines in human mobility. Data Mining and Knowledge Discovery 32, 787-829, https://link.springer.com/article/10.1007/s10618-017-0548-4
    """
    # if 'uid' column in not present in the TrajDataFrame
    if constants.UID not in traj.columns:
        return pd.DataFrame(pd.Series([_waiting_times_individual(traj)]), columns=[sys._getframe().f_code.co_name])
    
    if show_progress:
        df = traj.groupby(constants.UID).progress_apply(lambda x: _waiting_times_individual(x))
    else:
        df = traj.groupby(constants.UID).apply(lambda x: _waiting_times_individual(x))
    
    df = pd.DataFrame(df).reset_index().rename(columns={0: sys._getframe().f_code.co_name})
    
    if merge:
        wl_list =[]
        for x in df.waiting_times:
            wl_list.extend(x)
        return wl_list
    
    return df


def _number_of_locations_individual(traj):
    """
    Compute the number of visited locations of a single individual given their TrajDataFrame.

    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individual.
    
    Returns
    -------
    int
        number of distinct locations visited by the individual.
    """
    n_locs = len(traj.groupby([constants.LATITUDE, constants.LONGITUDE]).groups)
    return n_locs


def number_of_locations(traj, show_progress=True):
    """Number of distinct locations.
    
    Compute the number of distinct locations visited by a set of individuals in a TrajDataFrame [GHB2008]_.
    
    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individuals
    
    show_progress : boolean, optional
        if True, show a progress bar. The default is True.
    
    Returns
    -------
    pandas DataFrame
        the number of distinct locations visited by the individuals.
    
    Examples
    --------
    >>> import skmob
    >>> from skmob.measures.individual import number_of_locations
    >>> url = "https://snap.stanford.edu/data/loc-brightkite_totalCheckins.txt.gz"
    >>> df = pd.read_csv(url, sep='\\t', header=0, nrows=100000, 
                 names=['user', 'check-in_time', 'latitude', 'longitude', 'location id'])
    >>> tdf = skmob.TrajDataFrame(df, latitude='latitude', longitude='longitude', datetime='check-in_time', user_id='user')
    >>> nl_df = number_of_locations(tdf)
    >>> print(nl_df.head())
       uid  number_of_locations
    0    0                  542
    1    1                   97
    2    2                  460
    3    3                  614
    4    4                  216
    """
    # if 'uid' column in not present in the TrajDataFrame
    if constants.UID not in traj.columns:
        return pd.DataFrame([_number_of_locations_individual(traj)], columns=[sys._getframe().f_code.co_name])
    
    if show_progress:
        df = traj.groupby(constants.UID).progress_apply(lambda x: _number_of_locations_individual(x))
    else:
        df = traj.groupby(constants.UID).apply(lambda x: _number_of_locations_individual(x))
    return pd.DataFrame(df).reset_index().rename(columns={0: sys._getframe().f_code.co_name})


def _home_location_individual(traj, start_night='22:00', end_night='07:00'):
    """
    Compute the home location of a single individual given their TrajDataFrame.
    
    Parameters
    ----------
    traj : TrajDataFrame
        the trajectory of the individual.
    
    start_night : str, optional
        the starting time of the night (format HH:MM). The default is '22:00'.
        
    end_night : str, optional
        the ending time for the night (format HH:MM). The default is '07:00'.
    
    Returns
    -------
    tuple
        the latitude and longitude coordinates of the individual's home location 
    """
    night_visits = traj.set_index(pd.DatetimeIndex(traj.datetime)).between_time(start_night, end_night)
    if len(night_visits) != 0:
        lat, lng = night_visits.groupby([constants.LATITUDE, constants.LONGITUDE]).count().sort_values(by=constants.DATETIME, ascending=False).iloc[0].name
    else:
        lat, lng = traj.groupby([constants.LATITUDE, constants.LONGITUDE]).count().sort_values(by=constants.DATETIME, ascending=False).iloc[0].name
    home_coords = (lat, lng)
    return home_coords


def home_location(traj, start_night='22:00', end_night='07:00', show_progress=True):
    """Home location.
    
    Compute the home location of a set of individuals in a TrajDataFrame. The home location :math:`h(u)` of an individual :math:`u` is defined as the location :math:`u` visits the most during nighttime [CBTDHVSB2012]_ [PSO2012]_: 
    
    .. math:: 
        h(u) = \\arg\max_{i} |\{r_i | t(r_i) \in [t_{startnight}, t_{endnight}] \}|
    
    where :math:`r_i` is a location visited by :math:`u`, :math:`t(r_i)` is the time when :math:`u` visited :math:`r_i`, and :math:`t_{startnight}` and :math:`t_{endnight}` indicates the times when nighttime starts and ends, respectively.

    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individuals.
    
    start_night : str, optional
        the starting time of the night (format HH:MM). The default is '22:00'.
        
    end_night : str, optional
        the ending time for the night (format HH:MM). The default is '07:00'.
    
    show_progress : boolean, optional
        if True, show a progress bar. The default is True.
    
    Returns
    -------
    pandas DataFrame
        the home location, as a :math:`(latitude, longitude)` pair, of the individuals.
    
    Examples
    --------
    >>> import skmob
    >>> from skmob.measures.individual import home_location
    >>> url = "https://snap.stanford.edu/data/loc-brightkite_totalCheckins.txt.gz"
    >>> df = pd.read_csv(url, sep='\\t', header=0, nrows=100000, 
                 names=['user', 'check-in_time', 'latitude', 'longitude', 'location id'])
    >>> tdf = skmob.TrajDataFrame(df, latitude='latitude', longitude='longitude', datetime='check-in_time', user_id='user')
    >>> hl_df = home_location(tdf)
    >>> print(hl_df.head())
       uid        lat         lng
    0    0  39.891077 -105.068532
    1    1  37.630490 -122.411084
    2    2  39.739154 -104.984703
    3    3  37.748170 -122.459192
    4    4  60.180171   24.949728
    
    References
    ----------
    .. [CBTDHVSB2012] Csáji, B. C., Browet, A., Traag, V. A., Delvenne, J.-C., Huens, E., Van Dooren, P., Smoreda, Z. & Blondel, V. D. (2012) Exploring the Mobility of Mobile Phone Users. Physica A: Statistical Mechanics and its Applications 392(6), 1459-1473, https://www.sciencedirect.com/science/article/pii/S0378437112010059
    .. [PSO2012] Phithakkitnukoon, S., Smoreda, Z. & Olivier, P. (2012) Socio-geography of human mobility: A study using longitudinal mobile phone data. PLOS ONE 7(6): e39253. https://doi.org/10.1371/journal.pone.0039253
    
    See Also
    --------
    max_distance_from_home
    """
    # if 'uid' column in not present in the TrajDataFrame
    if constants.UID not in traj.columns:
        return pd.DataFrame([_home_location_individual(traj, start_night=start_night, end_night=end_night)], columns=[constants.LATITUDE, constants.LONGITUDE])
    
    if show_progress:
        df = traj.groupby(constants.UID).progress_apply(lambda x: _home_location_individual(x, start_night=start_night, end_night=end_night))
    else:
        df = traj.groupby(constants.UID).apply(lambda x: _home_location_individual(x, start_night=start_night, end_night=end_night))
    return df.apply(pd.Series).reset_index().rename(columns={0: constants.LATITUDE, 1: constants.LONGITUDE})


def _max_distance_from_home_individual(traj, start_night='22:00', end_night='07:00'):
    """
    Compute the maximum distance from home traveled by a single individual, given their TrajDataFrame.

    Parameters
    ----------
    traj : TrajDataFrame
        the trajectory of the individual.
    
    start_night : str, optional
        the starting time of the night (format HH:MM). The default is '22:00'.
        
    end_night : str, optional
        the ending time for the night (format HH:MM). The default is '07:00'.
    
    Returns
    -------
    float
        the maximum distance from home traveled by the individual.
    """    
    lats_lngs = traj.sort_values(by=constants.DATETIME)[[constants.LATITUDE, constants.LONGITUDE]].values
    home = home_location(traj, start_night=start_night, end_night=end_night, show_progress=False).iloc[0]
    home_lat, home_lng = home[constants.LATITUDE], home[constants.LONGITUDE]
    lengths = np.array([getDistanceByHaversine((lat, lng), (home_lat, home_lng)) for i, (lat, lng) in enumerate(lats_lngs)])
    return lengths.max()


def max_distance_from_home(traj, start_night='22:00', end_night='07:00', show_progress=True):
    """Maximum distance from home.
    
    Compute the maximum distance (in kilometers) traveled from their home location by a set of individuals in a TrajDataFrame. The maximum distance from home :math:`dh_{max}(u)` of an individual :math:`u` is defined as [CM2015]_:
    
    .. math:: 
        dh_{max}(u) = \max\limits_{1 \leq i \lt j \lt n_u} dist(r_i, h(u))
    
    where :math:`n_u` is the number of points recorded for :math:`u`, :math:`r_i` is a location visited by :math:`u` described as a :math:`(latitude, longitude)` pair, :math:`h(u)` is the home location of :math:`u`, and :math:`dist` is the geographic distance between two points.
    
    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individuals.
    
    start_night : str, optional
        the starting time of the night (format HH:MM). The default is '22:00'.
        
    end_night : str, optional
        the ending time for the night (format HH:MM). The default is '07:00'.
    
    show_progress : boolean, optional
        if True, show a progress bar. The default is True.
    
    Returns
    -------
    pandas DataFrame
        the maximum distance from home of the individuals.
    
    Examples
    --------
    >>> import skmob
    >>> from skmob.measures.individual import max_distance_from_home
    >>> url = "https://snap.stanford.edu/data/loc-brightkite_totalCheckins.txt.gz"
    >>> df = pd.read_csv(url, sep='\\t', header=0, nrows=100000, 
                 names=['user', 'check-in_time', 'latitude', 'longitude', 'location id'])
    >>> tdf = skmob.TrajDataFrame(df, latitude='latitude', longitude='longitude', datetime='check-in_time', user_id='user')
    >>> dh_max = max_distance_from_home(tdf)
    >>> print(df_max.head())
       uid  max_distance_from_home
    0    0            11286.942949
    1    1            12800.547682
    2    2            11282.748348
    3    3            12799.754644
    4    4            15512.788707

    References
    ----------
    .. [CM2015] Canzian, L. & Musolesi, M. (2015) Trajectories of depression: unobtrusive monitoring of depressive states by means of smartphone mobility traces analysis. Proceedings of the 2015 ACM International Joint Conference on Pervasive and Ubiquitous Computing, 1293-1304, https://dl.acm.org/citation.cfm?id=2805845
    
    See Also
    --------
    maximum_distance, home_location
    """
    # if 'uid' column in not present in the TrajDataFrame
    if constants.UID not in traj.columns:
        return pd.DataFrame([_max_distance_from_home_individual(traj, 
                                                                start_night=start_night, 
                                                                end_night=end_night)], columns=[sys._getframe().f_code.co_name])
    
    if show_progress:
        df = traj.groupby(constants.UID).progress_apply(lambda x: _max_distance_from_home_individual(x, start_night=start_night, end_night=end_night))
    else:
        df = traj.groupby(constants.UID).apply(lambda x: _max_distance_from_home_individual(x, start_night=start_night, end_night=end_night))
    return pd.DataFrame(df).reset_index().rename(columns={0: sys._getframe().f_code.co_name})


def number_of_visits(traj, show_progress=True):
    """Number of visits.
    
    Compute the number of visits (i.e., data points) for each individual in a TrajDataFrame.

    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individuals.
    
    show_progress : boolean, optional
        if True, show a progress bar. The default is True.
    
    Returns
    -------
    pandas DataFrame
        the number of visits or points per each individual.
    
    Examples
    --------
    >>> import skmob
    >>> from skmob.measures.individual import number_of_visits
    >>> url = "https://snap.stanford.edu/data/loc-brightkite_totalCheckins.txt.gz"
    >>> df = pd.read_csv(url, sep='\\t', header=0, nrows=100000, 
                 names=['user', 'check-in_time', 'latitude', 'longitude', 'location id'])
    >>> tdf = skmob.TrajDataFrame(df, latitude='latitude', longitude='longitude', datetime='check-in_time', user_id='user')
    >>> num_v_df = number_of_visits(tdf)
    >>> print(num_v_df.head())
       uid  number_of_visits
    0    0              2099
    1    1              1210
    2    2              2100
    3    3              1807
    4    4               779
    """
    # if 'uid' column in not present in the TrajDataFrame
    if constants.UID not in traj.columns:
        return len(traj)
    
    if show_progress:
        df = traj.groupby(constants.UID).progress_apply(lambda x: len(x))
    else:
        df = traj.groupby(constants.UID).apply(lambda x: len(x))
    return pd.DataFrame(df).reset_index().rename(columns={0: sys._getframe().f_code.co_name})


def _location_frequency_individual(traj, normalize=True,
                                   location_columns=[constants.LATITUDE, constants.LONGITUDE]):
    """
    Compute the visitation frequency of each location for a single individual given their TrajDataFrame.
    
    Parameters
    ----------
    traj : TrajDataFrame
        the trajectory of the individual.
    
    normalize : boolean, optional
        if True, compute the ratio of visits, otherwise the row count of visits to each location. The default is True.
    
    location_columns : list, optional
        the name of the column(s) indicating the location. The default is [constants.LATITUDE, constants.LONGITUDE].
    
    Returns
    -------
    pandas DataFrame
        the location frequency of each location of the individual. 
    """
    freqs = traj.groupby(location_columns).count()[constants.DATETIME].sort_values(ascending=False)
    if normalize:
        freqs /= freqs.sum()
    return freqs


def location_frequency(traj, normalize=True, as_ranks=False, show_progress=True,
                       location_columns=[constants.LATITUDE, constants.LONGITUDE]):
    """Location frequency.
    
    Compute the visitation frequency of each location, for a set of individuals in a TrajDataFrame. Given an individual :math:`u`, the visitation frequency of a location :math:`r_i` is the number of visits to that location by :math:`u`. The visitation frequency :math:`f(r_i)` of location :math:`r_i` is also defined in the literaure as the probability of visiting location :math:`r_i` by :math:`u` [SKWB2010]_ [PF2018]_:
    
    .. math::
        f(r_i) = \\frac{n(r_i)}{n_u}
        
    where :math:`n(r_i)` is the number of visits to location :math:`r_i` by :math:`u`, and :math:`n_u` is the total number of data points in :math:`u`'s trajectory.

    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individuals.
    
    normalize : boolean, optional
        if True, the number of visits to a location by an individual is computed as probability, i.e., divided by the individual's total number of visits. The default is True.
    
    as_ranks : boolean, optional
        if True, return a list where element :math:`i` indicates the average visitation frequency of the :math:`i`-th most frequent location. The default is False.
   
    show_progress : boolean, optional
        if True, show a progress bar. The default is True.
        
    location_columns : list, optional
        the name of the column(s) indicating the location. The default is [constants.LATITUDE, constants.LONGITUDE].
    
    Returns
    -------
    pandas DataFrame or list
        the location frequency for each location for each individual, or the ranks list for each individual.
    
    Examples
    --------
    >>> import skmob
    >>> from skmob.measures.individual import location_frequency
    >>> url = "https://snap.stanford.edu/data/loc-brightkite_totalCheckins.txt.gz"
    >>> df = pd.read_csv(url, sep='\\t', header=0, nrows=100000, 
                 names=['user', 'check-in_time', 'latitude', 'longitude', 'location id'])
    >>> tdf = skmob.TrajDataFrame(df, latitude='latitude', longitude='longitude', datetime='check-in_time', user_id='user')
    >>> lf_df = location_frequency(tdf, normalize=False).reset_index()
    >>> print(lf_df.head())
       uid        lat         lng  location_frequency
    0    0  39.762146 -104.982480                 214
    1    0  39.891077 -105.068532                 137
    2    0  39.739154 -104.984703                 126
    3    0  39.891586 -105.068463                  72
    4    0  39.827022 -105.143191                  53
    >>> lf_df = location_frequency(tdf, normalize=True).reset_index() # frequencies ad probabilities
    >>> print(lf_df.head())
           uid        lat         lng  location_frequency
    0    0  39.762146 -104.982480            0.101953
    1    0  39.891077 -105.068532            0.065269
    2    0  39.739154 -104.984703            0.060029
    3    0  39.891586 -105.068463            0.034302
    4    0  39.827022 -105.143191            0.025250
    >>> ranks = location_frequency(tdf, as_ranks=True) # as rank list
    >>> print(ranks[:10])
    [0.26774954912290716, 0.12699129836809203, 0.07090642778490935, 0.04627646190564675, 0.03657120208870922, 0.029353331229094993, 0.025050267239164755, 0.020284764933447663, 0.018437443393907686, 0.01656729815097415]
    
    See Also
    --------
    visits_per_location
    """
    # TrajDataFrame without 'uid' column
    if constants.UID not in traj.columns: 
        df = pd.DataFrame(_location_frequency_individual(traj, location_columns=location_columns))
        return df.reset_index()
    
    # TrajDataFrame with a single user
    n_users = len(traj[constants.UID].unique())
    if n_users == 1: # if there is only one user in the TrajDataFrame
        df = pd.DataFrame(_location_frequency_individual(traj, location_columns=location_columns))
        return df.reset_index()
    
    # TrajDataFrame with multiple users
    if show_progress:
        df = pd.DataFrame(traj.groupby(constants.UID)
                          .progress_apply(lambda x: _location_frequency_individual(x, normalize=normalize, location_columns=location_columns)))
    else:
        df = pd.DataFrame(traj.groupby(constants.UID)
                          .apply(lambda x: _location_frequency_individual(x, normalize=normalize, location_columns=location_columns)))
    
    df = df.rename(columns={constants.DATETIME: 'location_frequency'})
    
    if as_ranks:
        ranks = [[] for i in range(df.groupby('uid').count().max().location_frequency)]
        for i, group in df.groupby('uid'):
            for j, (index, row) in enumerate(group.iterrows()):
                ranks[j].append(row.location_frequency)
        ranks = [np.mean(rr) for rr in ranks]
        return ranks
    
    return df


def _individual_mobility_network_individual(traj, self_loops=False):
    """
    Compute the individual mobility network of a single individual given their TrajDataFrame.
    
    Parameters
    -----------
    traj : TrajDataFrame
        the trajectory of the individual.
    
    self_loops : boolean, optional
        if True adds self loops also. The default is False.
    
    Returns
    -------
    pandas DataFrame
        the individual mobility network of the individual.
    """
    loc2loc2weight = defaultdict(lambda: defaultdict(lambda: 0))
    traj = traj.sort_values(by=constants.DATETIME)
    lats_lngs = traj[[constants.LATITUDE, constants.LONGITUDE]].values

    i = 1
    for lat, lng in lats_lngs[1:]:
        prev = tuple(lats_lngs[i - 1])
        current = (lat, lng)
        if prev != current:
            loc2loc2weight[prev][current] += 1
        elif self_loops:
            loc2loc2weight[prev][current] += 1
        else:
            pass
        i += 1

    rows = []
    for loc1, loc2weight in loc2loc2weight.items():
        for loc2, weight in loc2weight.items():
            rows.append([loc1[0], loc1[1], loc2[0], loc2[1], weight])
    return pd.DataFrame(rows, columns=[constants.LATITUDE + '_origin', constants.LONGITUDE + '_origin',
                                       constants.LATITUDE + '_dest', constants.LONGITUDE + '_dest', 'n_trips'])


def individual_mobility_network(traj, self_loops=False, show_progress=True):
    """Individual Mobility Network.
    
    Compute the individual mobility network of a set of individuals in a TrajDataFrame. An Individual Mobility Network (aka IMN) of an individual :math:`u` is a directed graph :math:`G_u=(V,E)`, where :math:`V` is the set of nodes and :math:`E` is the set of edges. Nodes indicate locations visisted by :math:`u`, and edges indicate trips between two locations by :math:`u`. On the edges the following function is defined:
    
    .. math::
        \omega: E \\rightarrow \mathbb{N} 
        
    which returns the weight of an edge, i.e., the number of travels performed by :math:`u` on that edge [RGNPPG2014]_ [BL2012]_ [SQBB2010]_.
    
    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individuals.
    
    self_loops : boolean, optional
        if True, adds self loops also. The default is False.
    
    show_progress : boolean, optional
        if True, show a progress bar. The default is True.
    
    Returns
    -------
    pandas DataFrame
        the individual mobility network of each individual.

    Warning
    -------
    The input TrajDataFrame must be sorted in ascending order by `datetime`.

    Examples
    --------
    >>> import skmob
    >>> from skmob.measures.individual import individual_mobility_network
    >>> url = "https://snap.stanford.edu/data/loc-brightkite_totalCheckins.txt.gz"
    >>> df = pd.read_csv(url, sep='\\t', header=0, nrows=100000, 
                 names=['user', 'check-in_time', 'latitude', 'longitude', 'location id'])
    >>> tdf = skmob.TrajDataFrame(df, latitude='latitude', longitude='longitude', datetime='check-in_time', user_id='user')
    >>> imn_df = individual_mobility_network(tdf)
    >>> print(imn_df.head())
       uid  lat_origin  lng_origin   lat_dest    lng_dest n_trips
    0    0   37.774929 -122.419415  37.600747 -122.382376       1
    1    0   37.600747 -122.382376  37.615223 -122.389979       1
    2    0   37.600747 -122.382376  37.580304 -122.343679       1
    3    0   37.615223 -122.389979  39.878664 -104.682105       1
    4    0   37.615223 -122.389979  37.580304 -122.343679       1

    References
    ----------
    .. [RGNPPG2014] Rinzivillo, S., Gabrielli, L., Nanni, M., Pappalardo, L., Pedreschi, D. & Giannotti, F. (2012) The purpose of motion: Learning activities from Individual Mobility Networks. Proceedings of the 2014 IEEE International Conference on Data Science and Advanced Analytics, 312-318, https://ieeexplore.ieee.org/document/7058090
    .. [BL2012] Bagrow, J. P. & Lin, Y.-R. (2012) Mesoscopic Structure and Social Aspects of Human Mobility. PLOS ONE 7(5): e37676. https://doi.org/10.1371/journal.pone.0037676
    """
    # if 'uid' column in not present in the TrajDataFrame
    if constants.UID not in traj.columns:
        return _individual_mobility_network_individual(traj)
    
    if show_progress:
        return traj.groupby(constants.UID).progress_apply(lambda x: _individual_mobility_network_individual(x,
                self_loops=self_loops)).reset_index().drop('level_1', axis=1)
    else:
        return traj.groupby(constants.UID).apply(lambda x: _individual_mobility_network_individual(x,
                self_loops=self_loops)).reset_index().drop('level_1', axis=1)


def _recency_rank_individual(traj):
    """
    Compute the recency rank of the locations of an individual given their TrajDataFrame.

    Parameters
    ----------
    traj : TrajDataFrame
        the trajectory of the individual.
    
    Returns
    -------
    pandas DataFrame
        the recency rank for each location of the individual.
    """
    traj = traj.sort_values(constants.DATETIME, ascending=False).drop_duplicates(subset=[constants.LATITUDE,
                                                                                         constants.LONGITUDE],
                                                                                 keep="first")
    traj['recency_rank'] = range(1, len(traj) + 1)
    return traj[[constants.LATITUDE, constants.LONGITUDE, 'recency_rank']]


def recency_rank(traj, show_progress=True):
    """Recency rank.
    
    Compute the recency rank of the locations of a set of individuals in a TrajDataFrame. The recency rank :math:`K_s(r_i)` of a location :math:`r_i` of an individual :math:`u` is :math:`K_s(r_i) = 1` if location :math:`r_i` is the last visited location, it is :math:`K_s(r_i) = 2` if :math:`r_i` is the second-last visited location, and so on [BDEM2015]_. 
    
    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individuals.
    
    show_progress : boolean, optional
        if True, show a progress bar. The default is True.
    
    Returns
    -------
    pandas DataFrame
        the recency rank for each location of the individuals.
    
    Warning
    -------
    The input TrajDataFrame must be sorted in ascending order by `datetime`.
    
    Examples
    --------
    >>> import skmob
    >>> from skmob.measures.individual import recency_rank
    >>> url = "https://snap.stanford.edu/data/loc-brightkite_totalCheckins.txt.gz"
    >>> df = pd.read_csv(url, sep='\\t', header=0, nrows=100000, 
                 names=['user', 'check-in_time', 'latitude', 'longitude', 'location id'])
    >>> tdf = skmob.TrajDataFrame(df, latitude='latitude', longitude='longitude', datetime='check-in_time', user_id='user')
    >>> rr_df = recency_rank(tdf)
    >>> print(rr_df.head())
                 lat         lng  recency_rank
    uid                                       
    0   0  39.891383 -105.070814             1
        1  39.891077 -105.068532             2
        2  39.750469 -104.999073             3
        3  39.752713 -104.996337             4
        4  39.752508 -104.996637             5

    References
    ----------
    .. [BDEM2015] Barbosa, H., de Lima-Neto, F. B., Evsukoff, A., Menezes, R. (2015) The effect of recency to human mobility, EPJ Data Science 4(21), https://epjdatascience.springeropen.com/articles/10.1140/epjds/s13688-015-0059-8
    
    See Also
    --------
    frequency_rank
    """
    # if 'uid' column in not present in the TrajDataFrame
    if constants.UID not in traj.columns:
        return _recency_rank_individual(traj)
    
    if show_progress:
        df = traj.groupby(constants.UID).progress_apply(lambda x: _recency_rank_individual(x))
    else:
        df = traj.groupby(constants.UID).apply(lambda x: _recency_rank_individual(x))
    return pd.DataFrame(df)


def _frequency_rank_individual(traj):
    """
    Compute the frequency rank of the locations of a single individual given their TrajDataFrame.
    
    Parameters
    ----------
    traj : TrajDataFrame
        the trajectory of the individual.
    
    Returns
    -------
    pandas DataFrame
        the frequency rank for each location of the individual.
    """
    traj = traj.groupby([constants.LATITUDE, constants.LONGITUDE]).count().sort_values(by=constants.DATETIME, ascending=False).reset_index()
    traj['frequency_rank'] = range(1, len(traj) + 1)
    return traj[[constants.LATITUDE, constants.LONGITUDE, 'frequency_rank']]


def frequency_rank(traj, show_progress=True):
    """Frequency rank.
    
    Compute the frequency rank of the locations of a set of individuals in a TrajDataFrame. The frequency rank :math:`K_f(r_i)` of a location :math:`r_i` of an individual :math:`u` is :math:`K_f(r_i) = 1` if location :math:`r_i` is the most visited location, it is :math:`K_f(r_i) = 2` if :math:`r_i` is the second-most visited location, and so on [BDEM2015]_.

    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individuals.
    
    show_progress : boolean, optional
        if True, show a progress bar. The default is True.
    
    Returns
    -------
    pandas DataFrame
        the frequency rank for each location of the individuals.
    
    Examples
    --------
    >>> import skmob
    >>> from skmob.measures.individual import frequency_rank
    >>> url = "https://snap.stanford.edu/data/loc-brightkite_totalCheckins.txt.gz"
    >>> df = pd.read_csv(url, sep='\\t', header=0, nrows=100000, 
                 names=['user', 'check-in_time', 'latitude', 'longitude', 'location id'])
    >>> tdf = skmob.TrajDataFrame(df, latitude='latitude', longitude='longitude', datetime='check-in_time', user_id='user')
    >>> fr_df = frequency_rank(tdf)
    >>> print(fr_df.head())
                 lat         lng  frequency_rank
    uid                                         
    0   0  39.762146 -104.982480               1
        1  39.891077 -105.068532               2
        2  39.739154 -104.984703               3
        3  39.891586 -105.068463               4
        4  39.827022 -105.143191               5

    See Also
    --------
    recency_rank
    """
    # if 'uid' column in not present in the TrajDataFrame
    if constants.UID not in traj.columns:
        return _frequency_rank_individual(traj)
    
    if show_progress:
        df = traj.groupby(constants.UID).progress_apply(lambda x: _frequency_rank_individual(x))
    else:
        df = traj.groupby(constants.UID).apply(lambda x: _frequency_rank_individual(x))
    return df



from ..utils import gislib, utils, constants
from ..core.trajectorydataframe import *
import numpy as np
import inspect

def filter(tdf, max_speed_kmh=500., include_loops=False, speed_kmh=5., max_loop=6, ratio_max=0.25):
    """Trajectory filtering.
    
    For each individual in a TrajDataFrame, filter out the trajectory points that are considered noise or outliers [Z2015]_.
    
    Parameters
    ----------
    tdf : TrajDataFrame
        the trajectories of the individuals.

    max_speed_kmh : float, optional
        delete a trajectory point if the speed (in km/h) from the previous point is higher than `max_speed_kmh`. The default is `500.0`.

    include_loops: boolean, optional
        If `True`, trajectory points belonging to short and fast "loops" are removed. Specifically, points are removed if within the next `max_loop` points the individual has come back to a distance (`ratio_max` * the maximum distance reached), AND the average speed (in km/h) is higher than `speed`. The default is `False`.
    
    speed : float, optional 
        the default is 5km/h (walking speed).

    max_loop : int, optional
        the default is `6`.

    ratio_max : float, optional
        the default is `0.25`.
    
    Returns
    -------
    TrajDataFrame
        the TrajDataFrame without the trajectory points that have been filtered out.
    
    Warnings
    --------
    if `include_loops` is `True`, the filter is very slow. Use only if raw data is really noisy.
    
    Examples
    --------
    >>> import skmob
    >>> import pandas as pd
    >>> from skmob.preprocessing import filtering
    >>> # read the trajectory data (GeoLife)
    >>> url = 'https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/geolife_sample.txt.gz'
    >>> df = pd.read_csv(url, sep=',', compression='gzip')
    >>> tdf = skmob.TrajDataFrame(df, latitude='lat', longitude='lon', user_id='user', datetime='datetime')
    >>> print(tdf.head())
             lat         lng            datetime  uid
    0  39.984094  116.319236 2008-10-23 05:53:05    1
    1  39.984198  116.319322 2008-10-23 05:53:06    1
    2  39.984224  116.319402 2008-10-23 05:53:11    1
    3  39.984211  116.319389 2008-10-23 05:53:16    1
    4  39.984217  116.319422 2008-10-23 05:53:21    1
    >>> # filter out all points with a speed (in km/h) from the previous point higher than 500 km/h
    >>> ftdf = filtering.filter(tdf, max_speed_kmh=500.)
    >>> print(ftdf.parameters)
    {'filter': {'function': 'filter', 'max_speed_kmh': 500.0, 'include_loops': False, 'speed_kmh': 5.0, 'max_loop': 6, 'ratio_max': 0.25}}
    >>> n_deleted_points = len(tdf) - len(ftdf) # number of deleted points
    >>> print(n_deleted_points)
    54
    
    References
    ----------
    .. [Z2015] Zheng, Y. (2015) Trajectory data mining: an overview. ACM Transactions on Intelligent Systems and Technology 6(3), https://dl.acm.org/citation.cfm?id=2743025
    """
    # Sort
    tdf = tdf.sort_by_uid_and_datetime()

    # Save function arguments and values in a dictionary
    frame = inspect.currentframe()
    args, _, _, arg_values = inspect.getargvalues(frame)
    arguments = dict([('function', filter.__name__)]+[(i, arg_values[i]) for i in args[1:]])

    groupby = []

    if utils.is_multi_user(tdf):
        groupby.append(constants.UID)
    if utils.is_multi_trajectory(tdf):
        groupby.append(constants.TID)

    if len(groupby) > 0:
        # Apply simplify trajectory to each group of points
        ftdf = tdf.groupby(groupby, group_keys=False).apply(_filter_trajectory, max_speed=max_speed_kmh,
                                                            include_loops=include_loops, speed=speed_kmh,
                                                             max_loop=max_loop, ratio_max=ratio_max)
    else:
        ftdf = _filter_trajectory(tdf, speed=speed_kmh, max_speed=max_speed_kmh,
                        max_loop=max_loop, ratio_max=ratio_max,
                        include_loops=include_loops)

    # TODO: remove the following line when issue #71 (Preserve the TrajDataFrame index during preprocessing operations) is solved.
    ftdf.reset_index(inplace=True, drop=True)

    ftdf.parameters = tdf.parameters
    ftdf.set_parameter(constants.FILTERING_PARAMS, arguments)
    return ftdf


def _filter_trajectory(tdf, max_speed, include_loops, speed, max_loop, ratio_max):
    # From dataframe convert to numpy matrix
    lat_lng_dtime_other = list(utils.to_matrix(tdf))
    columns_order = list(tdf.columns)

    trajectory = _filter_array(lat_lng_dtime_other, max_speed, include_loops, speed, max_loop, ratio_max)

    filtered_traj = nparray_to_trajdataframe(trajectory, utils.get_columns(tdf), {})
    # Put back to the original order
    filtered_traj = filtered_traj[columns_order]

    return filtered_traj


def _filter_array(lat_lng_dtime_other, max_speed, include_loops, speed, max_loop, ratio_max):
    """
    TODO: add a filter based on the acceleration

     Delete points from raw trajectory `data` if:

    1. The speed from previous point is > `max_speed` km/h

    2. Within the next `max_loop` points the user has come back
            of `ratio_max`% of the maximum distance reached, AND s/he travelled
            at a speed > `speed` km/h
     """
    distfunc = gislib.getDistance
    lX = len(lat_lng_dtime_other)
    i = 0

    while i < lX - 2:

        try:
            dt = utils.diff_seconds(lat_lng_dtime_other[i][2], lat_lng_dtime_other[i + 1][2])

        except IndexError:
            pass
        try:
            if distfunc(lat_lng_dtime_other[i][:2], lat_lng_dtime_other[i + 1][:2]) / dt * 3600. > max_speed:
                del lat_lng_dtime_other[i + 1]
                lX = len(lat_lng_dtime_other)
                continue

        except ZeroDivisionError:
            del lat_lng_dtime_other[i + 1]
            lX = len(lat_lng_dtime_other)
            continue

        if include_loops:
            ahead = min(max_loop, lX - i - 1)
            DrDt = np.array([[distfunc(lat_lng_dtime_other[i][:2], lat_lng_dtime_other[i + j][:2]),
                              utils.diff_seconds(lat_lng_dtime_other[i][2], lat_lng_dtime_other[i + j][2])] for j in range(1, ahead)])

            imax = np.argmax(DrDt, axis=0)[0]
            inside = imax + np.where(DrDt[imax:, 0] < DrDt[imax, 0] * ratio_max)[0]
            try:
                imin = inside[0]
            except IndexError:
                i += 1
                continue
            Dr, Dt = sum(DrDt[:imin, :])

            if Dr / Dt * 3600. > speed:
                del lat_lng_dtime_other[i + 1 + imax]
                lX = len(lat_lng_dtime_other)
            else:
                i += 1
        else:
            i += 1

    return lat_lng_dtime_other


import powerlaw
import pandas as pd
import numpy as np
from collections import defaultdict
import datetime
import math
from tqdm import tqdm
from ..utils import constants, utils, gislib
from scipy.sparse import lil_matrix
import random
import logging
import inspect
from ..core.trajectorydataframe import TrajDataFrame
from ..models.gravity import Gravity

earth_distance_km = gislib.getDistance

latitude = constants.LATITUDE
longitude = constants.LONGITUDE
date_time = constants.DATETIME
user_id = constants.UID


def compute_od_matrix(gravity_singly, spatial_tessellation, tile_id_column=constants.TILE_ID,
                      relevance_column=constants.RELEVANCE):
    """
    Compute a matrix :math:`M` where element :math:`M_{ij}` is the probability p_{ij} of moving between
    locations :math:`i` and location :math:`j`, where each location refers to a row in `spatial_tessellation`.

    Parameters
    ----------
    gravity_singly : object
        instance of class `collective.Gravity` with argument `gravity_type='singly constrained'`.

    spatial_tessellation : GeoDataFrame
        the spatial tessellation describing the division of the territory in locations.

    tile_id_column : str or int, optional
        column of in `spatial_tessellation` containing the identifier of the location/tile. The default value is constants.TILE_ID.

    relevance_column : str or int, optional
        column in `spatial_tessellation` containing the relevance of the location/tile.

    Returns
    -------
    od_matrix : numpy array
        two-dimensional numpy array with the trip probabilities for each origin-destination pair.
    """
    od_matrix = gravity_singly.generate(spatial_tessellation,
                                        tile_id_column=tile_id_column,
                                        tot_outflows_column=None,
                                        relevance_column=relevance_column,
                                        out_format='probabilities')
    return od_matrix


def populate_od_matrix(location, lats_lngs, relevances, gravity_singly):
    """
    Populate the origin-destination matrix with the probability to move from the location in input to all other locations in the spatial tessellation.
    
    Parameters
    ----------
    location : int
        the identifier of a location.

    lats_lngs : list or numpy array
        list of coordinates of the centroids of the tiles in a spatial tessellation.

    relevances : list or numpy array
        list of relevances of the tiles in a spatial tessellation.

    gravity_singly : object
        instance of class `collective.Gravity` with argument `gravity_type='singly constrained'`.
    
    Returns
    -------
        a numpy array of trip probabilities between the origin location and each destination.
    """
    ll_origin = lats_lngs[location]
    distances = np.array([earth_distance_km(ll_origin, l) for l in lats_lngs])

    scores = gravity_singly.compute_gravity_score(distances, relevances[location], relevances)
    return scores / sum(scores)


class EPR:

    def __init__(self, name='EPR model', rho=0.6, gamma=0.21, beta=0.8, tau=17, min_wait_time_minutes=20):

        self._name = name

        self._rho = rho
        self._gamma = gamma
        self._tau = tau
        self._beta = beta

        self._location2visits = defaultdict(int)
        self._od_matrix = None
        self._is_sparse = True
        self._spatial_tessellation = None
        self.lats_lngs = None
        self.relevances = None
        self._starting_loc = None
        self.gravity_singly = None

        # Minimum waiting time (in hours)
        self._min_wait_time = min_wait_time_minutes / 60.0  # minimum waiting time

        self._trajectories_ = []
        self._log_file = None

    @property
    def name(self):
        return self._name

    @property
    def rho(self):
        return self._rho

    @property
    def gamma(self):
        return self._gamma

    @property
    def tau(self):
        return self._tau

    @property
    def beta(self):
        return self._beta

    @property
    def min_wait_time(self):
        return self._min_wait_time

    @property
    def spatial_tessellation_(self):
        return self._spatial_tessellation

    @property
    def trajectories_(self):
        return self._trajectories_

    def _weighted_random_selection(self, current_location):
        """
        Select a random location given the agent's visitation frequency. Used by the return mechanism.
        
        Parameters
        ----------
        current_location : int
            identifier of a location.
            
        Returns
        -------
        int
            a location randomly chosen according to its relevance.
        """
        locations = np.fromiter(self._location2visits.keys(), dtype=int)
        weights = np.fromiter(self._location2visits.values(), dtype=float)

        # remove the current location
        currloc_idx = np.where(locations == current_location)[0][0]
        locations = np.delete(locations, currloc_idx)
        weights = np.delete(weights, currloc_idx)

        weights = weights / np.sum(weights)
        location = np.random.choice(locations, size=1, p=weights)
        return int(location[0])

    def _preferential_return(self, current_location):
        """
        Choose the location the agent returns to, according to the visitation frequency
        of the previously visited locations.
        
        Parameters
        ----------
        current_location : int
            the current location where the agent is.
            
        Returns
        -------
        int
            the identifier of the next location the agent moves to.
        """
        next_location = self._weighted_random_selection(current_location)
        if self._log_file is not None:
            logging.info('RETURN to %s (%s, %s)' % (next_location, self.lats_lngs[next_location]))
            logging.info('\t frequency = %s' % self._location2visits[next_location])
        return next_location

    def _preferential_exploration(self, current_location):
        """
        Choose the new location the agent explores, according to the probabilities
        in the origin-destination matrix.
        
        Parameters
        ----------
        current_location : int
            the identifier of the current location of the individual.
        
        Returns
        -------
        int
            the identifier of the new location the agent has to explore.
        """

        if self._is_sparse:
            prob_array = self._od_matrix.getrowview(current_location)
            if prob_array.nnz == 0:
                # if the row has been not populated
                weights = populate_od_matrix(current_location, self.lats_lngs, self.relevances, self.gravity_singly)
                self._od_matrix[current_location, :] = weights
            else:
                weights = prob_array.toarray()[0]
            locations = np.arange(len(self.lats_lngs))
            location = np.random.choice(locations, size=1, p=weights)[0]

        else:  # if the matrix is precomputed
            locations = np.arange(len(self._od_matrix[current_location]))
            weights = self._od_matrix[current_location]
            location = np.random.choice(locations, size=1, p=weights)[0]

        if self._log_file is not None:
            logging.info('EXPLORATION to %s (%s, %s)' % (location, self.lats_lngs[location]))

        return location

    def _get_trajdataframe(self, parameters):
        """
        Transform the trajectories list into a pandas DataFrame.
        
        Returns
        -------
        pandas DataFrame
            the trajectories of the agent.
        """
        df = pd.DataFrame(self._trajectories_, columns=[user_id, date_time, 'location'])
        df[[latitude, longitude]] = df.location.apply(lambda s: pd.Series({latitude: self.lats_lngs[s][0],
                                                                           longitude: self.lats_lngs[s][1]}))
        df = df.sort_values(by=[user_id, date_time]).drop('location', axis=1)
        return TrajDataFrame(df, parameters=parameters)

    def _choose_location(self):
        """
        Choose the next location to visit given the agent's current location.
        
        Returns
        -------
        int
            the identifier of the next location the agent has to visit.
        """
        n_visited_locations = len(self._location2visits)  # number of already visited locations

        if n_visited_locations == 0:
            self._starting_loc = self._preferential_exploration(self._starting_loc)
            return self._starting_loc

        agent_id, current_time, current_location = self._trajectories_[-1]  # the last visited location

        # choose a probability to return or explore
        p_new = random.uniform(0, 1)

        if (p_new <= self._rho * math.pow(n_visited_locations, -self._gamma) and n_visited_locations != \
                self._od_matrix.shape[0]) or n_visited_locations == 1:  # choose to return or explore
            # PREFERENTIAL EXPLORATION
            next_location = self._preferential_exploration(current_location)
            # TODO: remove the part below and exclude visited locations
            #  from the list of potential destinations in _preferential_exploration
            # while next_location in self._location2visits:
            #     next_location = self._preferential_exploration(current_location)
            return next_location

        else:
            # PREFERENTIAL RETURN
            next_location = self._preferential_return(current_location)
            return next_location

    def _time_generator(self):
        return powerlaw.Truncated_Power_Law(xmin=self.min_wait_time,
                                            parameters=[1. + self._beta, 1.0 / self._tau]).generate_random()[0]

    def _choose_waiting_time(self):
        """
        Choose the time (in hours) the agent has to wait before the next movement.
        
        Returns
        -------
        float
            the time the agent has to wait before the next movement.
        """
        time_to_wait = self._time_generator()
        return time_to_wait

    def generate(self, start_date, end_date, spatial_tessellation, gravity_singly={}, n_agents=1,
                 starting_locations=None, od_matrix=None,
                 relevance_column=constants.RELEVANCE,
                 random_state=None, log_file=None, show_progress=False):
        """
        Start the simulation of a set of agents at time `start_date` till time `end_date`.
        
        Parameters
        ----------
        start_date : datetime
            the starting date of the simulation, in "YYY/mm/dd HH:MM:SS" format.

        end_date : datetime
            the ending date of the simulation, in "YYY/mm/dd HH:MM:SS" format.

        spatial_tessellation : geopandas GeoDataFrame
            the spatial tessellation, i.e., a division of the territory in locations. 
        
        gravity_singly : {} or Gravity, optional
            the gravity model (singly constrained or doubly constrained) to use when generating the probability to move between two locations. The default is "{}".
        
        n_agents : int, optional
            the number of agents to generate. The default is 1.

        starting_locations : list or None, optional
            a list of integers, each identifying the location from which to start the simulation of each agent. Note that, if `starting_locations` is not None, its length must be equal to the value of `n_agents`, i.e., you must specify one starting location per agent. The default is None.
        
        od_matrix : numpy array or None, optional
            the origin destination matrix to use for deciding the movements of the agent. If `None`, it is computed "on the fly" during the simulation. The default is None.
        
        relevance_column : str, optional
            the name of the column in `spatial_tessellation` to use as relevance variable. The default is "relevance".
        
        random_state : int or None, optional
            if int, it is the seed used by the random number generator; if None, the random number generator is the RandomState instance used by np.random and random.random. The default is None.
        
        log_file : str or None, optional
            the name of the file where to write a log of the execution of the model. The logfile will contain all decisions (returns or explorations) made by the model. The default is None.
            
        show_progress : boolean, optional
            if True, show a progress bar. The default is False.
        
        Returns
        -------
        TrajDataFrame
            the synthetic trajectories generated by the model
        """
        if starting_locations is not None and len(starting_locations) < n_agents:
            raise IndexError("The number of starting locations is smaller than the number of agents.")

        if gravity_singly == {}:
            self.gravity_singly = Gravity(gravity_type='singly constrained')

        # Save function arguments and values in a dictionary
        frame = inspect.currentframe()
        args, _, _, arg_values = inspect.getargvalues(frame)
        parameters = dict([])
        parameters['model'] = {'class': self.__class__.__init__,
                               'generate': {i: arg_values[i] for i in args[1:] if i not in ['spatial_tessellation',
                                                                                            'od_matrix', 'log_file',
                                                                                            'starting_locations']}}

        # if specified, fix the random seeds to guarantee reproducibility of simulation
        if random_state is not None:
            random.seed(random_state)
            np.random.seed(random_state)

        if log_file is not None:
            self._log_file = log_file
            logging.basicConfig(format='%(message)s', filename=log_file, filemode='w', level=logging.INFO)

        # initialization of trajectories
        self._trajectories_ = []

        # setting of spatial tessellation
        num_locs = len(spatial_tessellation)
        self.lats_lngs = spatial_tessellation.geometry.apply(utils.get_geom_centroid, args=[True]).values
        if relevance_column is None:
            self.relevances = np.ones(num_locs)
        else:
            self.relevances = spatial_tessellation[relevance_column].fillna(0).values

        # initialization of od matrix
        if od_matrix is None:
            self._od_matrix = lil_matrix((num_locs, num_locs))
            self._is_sparse = True
        else:
            self._od_matrix = od_matrix
            self._is_sparse = False

        # for each agent
        loop = range(1, n_agents + 1)
        if show_progress:
            loop = tqdm(range(1, n_agents + 1))

        for agent_id in loop:
            self._location2visits = defaultdict(int)
            if starting_locations is None:
                self._starting_loc = np.random.choice(np.fromiter(range(num_locs), dtype=int), size=1)[0]
            else:
                self._starting_loc = starting_locations.pop()

            self._epr_generate_one_agent(agent_id, start_date, end_date)

        tdf = self._get_trajdataframe(parameters)
        return tdf

    def _epr_generate_one_agent(self, agent_id, start_date, end_date):

        current_date = start_date
        self._trajectories_.append((agent_id, current_date, self._starting_loc))
        self._location2visits[self._starting_loc] += 1

        waiting_time = self._choose_waiting_time()
        current_date += datetime.timedelta(hours=waiting_time)

        while current_date < end_date:
            next_location = self._choose_location()
            self._trajectories_.append((agent_id, current_date, next_location))
            self._location2visits[next_location] += 1

            waiting_time = self._choose_waiting_time()
            current_date += datetime.timedelta(hours=waiting_time)


class DensityEPR(EPR):
    """Density-EPR model.
    
    The d-EPR model of individual human mobility consists of the following mechanisms [PSRPGB2015]_ [PSR2016]_: 
    
    **Waiting time choice**. The waiting time :math:`\Delta t` between two movements of the agent is chosen randomly from the distribution :math:`P(\Delta t) \sim \Delta t^{−1 −\\beta} \exp(−\Delta t/ \\tau)`. Parameters :math:`\\beta` and :math:`\\tau` correspond to arguments `beta` and `tau` of the constructor, respectively. 
    
    **Action selection**. With probability :math:`P_{new}=\\rho S^{-\\gamma}`, where :math:`S` is the number of distinct locations previously visited by the agent, the agent visits a new location (Exploration phase), otherwise it returns to a previously visited location (Return phase). Parameters :math:`\\rho` and :math:`\\gamma` correspond to arguments `rho` and `gamma` of the constructor, respectively.
    
    **Exploration phase**. If the agent that is currently in location :math:`i` explores a new location, then the new location :math:`j \\neq i` is selected according to the gravity model with probability :math:`p_{ij} = \\frac{1}{N} \\frac{n_i n_j}{r_{ij}^2}`, where :math:`n_{i (j)}` is the location's relevance, that is, the probability of a population to visit location :math:`i(j)`, :math:`r_{ij}` is the geographic distance between :math:`i` and :math:`j`, and :math:`N = \sum_{i, j \\neq i} p_{ij}` is a normalization constant. The number of distinct locations visited, :math:`S`, is increased by 1.
    
    **Return phase**. If the individual returns to a previously visited location, such a location :math:`i` is chosen with probability proportional to the number of time the agent visited :math:`i`, i.e., :math:`\Pi_i = f_i`, where :math:`f_i` is the visitation frequency of location :math:`i`.
    
    Parameters
    ----------
    name : str, optional
        the name of the instantiation of the d-EPR model. The default value is "Density EPR model".

    rho : float, optional
        it corresponds to the parameter :math:`\\rho \in (0, 1]` in the Action selection mechanism :math:`P_{new} = \\rho S^{-\gamma}` and controls the agent's tendency to explore a new location during the next move versus returning to a previously visited location. The default value is :math:`\\rho = 0.6` [SKWB2010]_.

    gamma : float, optional
        it corresponds to the parameter :math:`\gamma` (:math:`\gamma \geq 0`) in the Action selection mechanism :math:`P_{new} = \\rho S^{-\gamma}` and controls the agent's tendency to explore a new location during the next move versus returning to a previously visited location. The default value is :math:`\gamma=0.21` [SKWB2010]_.

    beta : float, optional
        it corresponds to the parameter :math:`\\beta` of the waiting time distribution in the Waiting time choice mechanism. The default value is :math:`\\beta=0.8` [SKWB2010]_.

    tau : int, optional
        it corresponds to the parameter :math:`\\tau` of the waiting time distribution in the Waiting time choice mechanism. The default value is :math:`\\tau = 17`, expressed in hours [SKWB2010]_.

    min_wait_time_minutes : int
        minimum waiting time between two movements, in minutes.
    
    Attributes
    ----------
    name : str
        the name of the instantiation of the model.

    rho : float
        the input parameter :math:`\\rho`.

    gamma : float
        the input parameters :math:`\gamma`.
        
    beta : float
        the input parameter :math:`\\beta`. 

    tau : int
        the input parameter :math:`\\tau`.

    min_wait_time_minutes : int
        the input parameters `min_wait_time_minutes`.

    Examples
    --------
    >>> import skmob
    >>> import pandas as pd
    >>> import geopandas as gpd
    >>> from skmob.models.epr import DensityEPR
    >>> url = 'https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/NY_counties_2011.geojson'
    >>> tessellation = gpd.read_file(url)
    >>> start_time = pd.to_datetime('2019/01/01 08:00:00')
    >>> end_time = pd.to_datetime('2019/01/14 08:00:00')
    >>> depr = DensityEPR()
    >>> tdf = depr.generate(start_time, end_time, tessellation, relevance_column='population', n_agents=100, show_progress=True)
    >>> print(tdf.head())
       uid                   datetime        lat        lng
    0    1 2019-01-01 08:00:00.000000  42.780819 -76.823724
    1    1 2019-01-01 09:45:58.388540  42.728060 -77.775510
    2    1 2019-01-01 10:16:09.406408  42.780819 -76.823724
    3    1 2019-01-01 17:13:39.999037  42.852827 -77.299810
    4    1 2019-01-01 19:24:27.353379  42.728060 -77.775510
    >>> print(tdf.parameters)
    {'model': {'class': <function DensityEPR.__init__ at 0x7f548a49cf28>, 'generate': {'start_date': Timestamp('2019-01-01 08:00:00'), 'end_date': Timestamp('2019-01-14 08:00:00'), 'gravity_singly': {}, 'n_agents': 100, 'relevance_column': 'population', 'random_state': None, 'show_progress': True}}}
    
    References
    ----------
    .. [PSRPGB2015] Pappalardo, L., Simini, F. Rinzivillo, S., Pedreschi, D. Giannotti, F. & Barabasi, A. L. (2015) Returners and Explorers dichotomy in human mobility. Nature Communications 6, https://www.nature.com/articles/ncomms9166
    .. [PSR2016] Pappalardo, L., Simini, F. Rinzivillo, S. (2016) Human Mobility Modelling: exploration and preferential return meet the gravity model. Procedia Computer Science 83, https://www.sciencedirect.com/science/article/pii/S1877050916302216
    .. [SKWB2010] Song, C., Koren, T., Wang, P. & Barabasi, A.L. (2010) Modelling the scaling properties of human mobility. Nature Physics 6, 818-823, https://www.nature.com/articles/nphys1760
    
    See Also
    --------
    EPR, SpatialEPR, Ditras
    """

    def __init__(self, name='Density EPR model', rho=0.6, gamma=0.21, beta=0.8, tau=17, min_wait_time_minutes=20):

        super().__init__()
        self._name = name
        
    def generate(self, start_date, end_date, spatial_tessellation, gravity_singly={}, n_agents=1,
                 starting_locations=None, od_matrix=None, relevance_column=constants.RELEVANCE,
                 random_state=None, log_file=None, show_progress=False):
        """
        Start the simulation of a set of agents at time `start_date` till time `end_date`.
        
        Parameters
        ----------
        start_date : datetime
            the starting date of the simulation, in "YYY/mm/dd HH:MM:SS" format.

        end_date : datetime
            the ending date of the simulation, in "YYY/mm/dd HH:MM:SS" format.

        spatial_tessellation : geopandas GeoDataFrame
            the spatial tessellation, i.e., a division of the territory in locations. 
        
        gravity_singly : {} or Gravity, optional
            the gravity model (singly constrained or doubly constrained) to use when generating the probability to move between two locations (note, used by DensityEPR). The default is "{}".
        
        n_agents : int, optional
            the number of agents to generate. The default is 1.

        relevance_column : str, optional
            the name of the column in `spatial_tessellation` to use as relevance variable. The default is "relevance".

        starting_locations : list or None, optional
            a list of integers, each identifying the location from which to start the simulation of each agent. Note that, if `starting_locations` is not None, its length must be equal to the value of `n_agents`, i.e., you must specify one starting location per agent. The default is None.
        
        od_matrix : numpy array or None, optional
            the origin destination matrix to use for deciding the movements of the agent. If `None`, it is computed "on the fly" during the simulation. The default is None.
        
        random_state : int or None, optional
            if int, it is the seed used by the random number generator; if None, the random number generator is the RandomState instance used by np.random and random.random. The default is None.
        
        log_file : str or None, optional
            the name of the file where to write a log of the execution of the model. The logfile will contain all decisions (returns or explorations) made by the model. The default is None.
            
        show_progress : boolean, optional
            if True, show a progress bar. The default is False.
        
        Returns
        -------
        TrajDataFrame
            the synthetic trajectories generated by the model
        """
        return super().generate(start_date, end_date, spatial_tessellation, gravity_singly=gravity_singly, n_agents=n_agents, starting_locations=starting_locations, od_matrix=od_matrix, relevance_column=relevance_column, random_state=random_state, log_file=log_file, show_progress=show_progress)


class SpatialEPR(EPR):
    """Spatial-EPR model.
    
    The s-EPR model of individual human mobility consists of the following mechanisms [PSRPGB2015]_ [PSR2016]_ [SKWB2010]_:
    
    **Waiting time choice**. The waiting time :math:`\Delta t` between two movements of the agent is chosen randomly from the distribution :math:`P(\Delta t) \sim \Delta t^{−1 −\\beta} \exp(−\Delta t/ \\tau)`. Parameters :math:`\\beta` and :math:`\\tau` correspond to arguments `beta` and `tau` of the constructor, respectively. 
    
    **Action selection**. With probability :math:`P_{new}=\\rho S^{-\\gamma}`, where :math:`S` is the number of distinct locations previously visited by the agent, the agent visits a new location (Exploration phase), otherwise it returns to a previously visited location (Return phase). Parameters :math:`\\rho` and :math:`\\gamma` correspond to arguments `rho` and `gamma` of the constructor, respectively.
    
    **Exploration phase**. If the agent that is currently in location :math:`i` explores a new location, then the new location :math:`j \\neq i` is selected according to the distance from the current location :math:`p_{ij} = \\frac{1}{r_{ij}^2}`, where :math:`r_{ij}` is the geographic distance between :math:`i` and :math:`j`. The number of distinct locations visited, :math:`S`, is increased by 1.
    
    **Return phase**. If the individual returns to a previously visited location, such a location :math:`i` is chosen with probability proportional to the number of time the agent visited :math:`i`, i.e., :math:`\Pi_i = f_i`, where :math:`f_i` is the visitation frequency of location :math:`i`.
    
    .. image:: https://media.springernature.com/full/springer-static/image/art%3A10.1038%2Fnphys1760/MediaObjects/41567_2010_Article_BFnphys1760_Fig2_HTML.jpg?as=webp
    Starting at time :math:`t` from the configuration shown in the left panel, indicating that the user visited previously :math:`S=4` locations with frequency :math:`f_i` that is proportional to the size of circles drawn at each location, at time :math:`t + \Delta t` (with :math:`Delta t` drawn from the :math:`P(\Delta t)` fat-tailed distribution) the user can either visit a new location at distance :math:`\Delta r` from his/her present location, or return to a previously visited location with probability :math:`P_{ret}=\\rho S^{-\\gamma}`, where the next location will be chosen with probability :math:`\Pi_i=f_i` (preferential return; lower panel). Figure from [SKWB2010]_.
    
    Parameters
    ----------
    name : str, optional
        the name of the instantiation of the s-EPR model. The default value is "Spatial EPR model".

    rho : float, optional
        it corresponds to the parameter :math:`\\rho \in (0, 1]` in the Action selection mechanism :math:`P_{new} = \\rho S^{-\gamma}` and controls the agent's tendency to explore a new location during the next move versus returning to a previously visited location. The default value is :math:`\\rho = 0.6` [SKWB2010]_.

    gamma : float, optional
        it corresponds to the parameter :math:`\gamma` (:math:`\gamma \geq 0`) in the Action selection mechanism :math:`P_{new} = \\rho S^{-\gamma}` and controls the agent's tendency to explore a new location during the next move versus returning to a previously visited location. The default value is :math:`\gamma=0.21` [SKWB2010]_.

    beta : float, optional
        it corresponds to the parameter :math:`\\beta` of the waiting time distribution in the Waiting time choice mechanism. The default value is :math:`\\beta=0.8` [SKWB2010]_.

    tau : int, optional
        it corresponds to the parameter :math:`\\tau` of the waiting time distribution in the Waiting time choice mechanism. The default value is :math:`\\tau = 17`, expressed in hours [SKWB2010]_.

    min_wait_time_minutes : int
        the input parameters `min_wait_time_minutes`
    
    Attributes
    ----------
    name : str
        the name of the instantiation of the model.

    rho : float
        the input parameter :math:`\\rho`.

    gamma : float
        the input parameters :math:`\gamma`.
        
    beta : float
        the input parameter :math:`\\beta`. 

    tau : int
        the input parameter :math:`\\tau`. 

    min_wait_time_minutes : int
        the input parameters `min_wait_time_minutes`.
    
    Examples
    --------
    >>> import skmob
    >>> import pandas as pd
    >>> import geopandas as gpd
    >>> from skmob.models.epr import SpatialEPR
    >>> url = 'https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/NY_counties_2011.geojson'
    >>> tessellation = gpd.read_file(url)
    >>> start_time = pd.to_datetime('2019/01/01 08:00:00')
    >>> end_time = pd.to_datetime('2019/01/14 08:00:00')
    >>> sepr = SpatialEPR()
    >>> tdf = sepr.generate(start_time, end_time, tessellation, n_agents=100, show_progress=True)
    >>> print(tdf.head())
       uid                   datetime        lat        lng
    0    1 2019-01-01 08:00:00.000000  42.267915 -77.383591
    1    1 2019-01-01 13:06:13.973868  42.633510 -77.105324
    2    1 2019-01-01 14:17:41.188668  42.452018 -76.473618
    3    1 2019-01-01 14:49:30.896248  42.633510 -77.105324
    4    1 2019-01-01 15:10:54.133150  43.382528 -78.230656
    >>> print(tdf.parameters)
    {'model': {'class': <function SpatialEPR.__init__ at 0x7f548a49e048>, 'generate': {'start_date': Timestamp('2019-01-01 08:00:00'), 'end_date': Timestamp('2019-01-14 08:00:00'), 'gravity_singly': {}, 'n_agents': 100, 'relevance_column': None, 'random_state': None, 'show_progress': True}}}    
    
    See Also
    --------
    EPR, DensityEPR, Ditras        
    """

    def __init__(self, name='Spatial EPR model', rho=0.6, gamma=0.21, beta=0.8, tau=17, min_wait_time_minutes=20):

        super().__init__()
        self._name = name

    def generate(self, start_date, end_date, spatial_tessellation, gravity_singly={}, n_agents=1,
                 starting_locations=None, od_matrix=None, random_state=None, log_file=None, show_progress=False):
        """
        Start the simulation of a set of agents at time `start_date` till time `end_date`.
        
        Parameters
        ----------
        start_date : datetime
            the starting date of the simulation, in "YYY/mm/dd HH:MM:SS" format.

        end_date : datetime
            the ending date of the simulation, in "YYY/mm/dd HH:MM:SS" format.

        spatial_tessellation : geopandas GeoDataFrame
            the spatial tessellation, i.e., a division of the territory in locations. 
        
        gravity_singly : {} or Gravity, optional
            the gravity model (singly constrained or doubly constrained) to use when generating the probability to move between two locations (note, used by DensityEPR). The default is "{}".
        
        n_agents : int, optional
            the number of agents to generate. The default is 1.

        starting_locations : list or None, optional
            a list of integers, each identifying the location from which to start the simulation of each agent. Note that, if `starting_locations` is not None, its length must be equal to the value of `n_agents`, i.e., you must specify one starting location per agent. The default is None.
        
        od_matrix : numpy array or None, optional
            the origin destination matrix to use for deciding the movements of the agent. If `None`, it is computed "on the fly" during the simulation. The default is None.
        
        random_state : int or None, optional
            if int, it is the seed used by the random number generator; if None, the random number generator is the RandomState instance used by np.random and random.random. The default is None.
        
        log_file : str or None, optional
            the name of the file where to write a log of the execution of the model. The logfile will contain all decisions (returns or explorations) made by the model. The default is None.
            
        show_progress : boolean, optional
            if True, show a progress bar. The default is False.
        
        Returns
        -------
        TrajDataFrame
            the synthetic trajectories generated by the model
        """
        return super().generate(start_date, end_date, spatial_tessellation, gravity_singly=gravity_singly, n_agents=n_agents, starting_locations=starting_locations, od_matrix=od_matrix, relevance_column=None, random_state=random_state, log_file=log_file, show_progress=show_progress)


class Ditras(EPR):
    """Ditras modelling framework.
    
    The DITRAS (DIary-based TRAjectory Simulator) modelling framework to simulate the spatio-temporal patterns of human mobility [PS2018]_. DITRAS consists of two phases: 
    
    **Mobility Diary Generation**. In the first phase, DITRAS generates a *mobility diary* which captures the temporal patterns of human mobility. 
    
    **Trajectory Generation**. In the second phase, DITRAS transforms the mobility diary into a mobility trajectory which captures the spatial patterns of human movements. 

    .. image:: https://raw.githubusercontent.com/jonpappalord/DITRAS/master/DITRAS_schema.png
    
    **Outline of the DITRAS framework**. DITRAS combines two probabilistic models: a diary generator (e.g., :math:`MD(t)`) and trajectory generator (e.g., d-EPR). The diary generator produces a mobility diary :math:`D`. The mobility diary :math:`D` is the input of the trajectory generator together with a weighted spatial tessellation of the territory :math:`L`. From :math:`D` and :math:`L` the trajectory generator produces a synthetic mobility trajectory :math:`S`.

    Parameters
    ----------
    diary_generator : MarkovDiaryGenerator
        the diary generator to use for generating the diary.
    
    name : str, optional
        the name of the instantiation of the Ditras model. The default value is "Ditras".

    rho : float, optional
        it corresponds to the parameter :math:`\\rho \in (0, 1]` in the Action selection mechanism  of the DensityEPR model :math:`P_{new} = \\rho S^{-\gamma}` and controls the agent's tendency to explore a new location during the next move versus returning to a previously visited location. The default value is :math:`\\rho = 0.6` [SKWB2010]_.

    gamma : float, optional
        it corresponds to the parameter :math:`\gamma` (:math:`\gamma \geq 0`) in the Action selection mechanism of the DensityEPR model :math:`P_{new} = \\rho S^{-\gamma}` and controls the agent's tendency to explore a new location during the next move versus returning to a previously visited location. The default value is :math:`\gamma=0.21` [SKWB2010]_.

    Attributes
    ----------
    diary_generator : MarkovDiaryGenerator
        the diary generator to use for generating the diary [PS2018]_.
    
    name : str
        the name of the instantiation of the model.

    rho : float
        the input parameter :math:`\\rho`.

    gamma : float
        the input parameters :math:`\gamma`.

    Examples
    --------
    >>> import skmob
    >>> from skmob.models.epr import Ditras
    >>> from skmob.models.markov_diary_generator import MarkovDiaryGenerator
    >>> from skmob.preprocessing import filtering, compression, detection, clustering
    >>> 
    >>> # load and preprocess data to train the MarkovDiaryGenerator
    >>> url = 'https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/geolife_sample.txt.gz'
    >>> df = pd.read_csv(url, sep=',', compression='gzip')
    >>> tdf = skmob.TrajDataFrame(df, latitude='lat', longitude='lon', user_id='user', datetime='datetime')
    >>> ctdf = compression.compress(tdf)
    >>> stdf = detection.stops(ctdf)
    >>> cstdf = clustering.cluster(stdf)
    >>> 
    >>> # instantiate and train the MarkovDiaryGenerator
    >>> mdg = MarkovDiaryGenerator()
    >>> mdg.fit(cstdf, 2, lid='cluster')
    >>> 
    >>> # set start time, end time and tessellation for the simulation
    >>> start_time = pd.to_datetime('2019/01/01 08:00:00')
    >>> end_time = pd.to_datetime('2019/01/14 08:00:00')
    >>> tessellation = gpd.GeoDataFrame.from_file("data/NY_counties_2011.geojson")
    >>> 
    >>> # instantiate the model
    >>> ditras = Ditras(mdg)
    >>> 
    >>> # run the model
    >>> ditras_tdf = ditras.generate(start_time, end_time, tessellation, relevance_column='population',
                    n_agents=3, od_matrix=None, show_progress=True)
    >>> print(ditras_tdf.head())
       uid            datetime        lat        lng
    0    1 2019-01-01 08:00:00  43.382528 -78.230656
    1    1 2019-01-02 03:00:00  43.309133 -77.680414
    2    1 2019-01-02 23:00:00  43.382528 -78.230656
    3    1 2019-01-03 10:00:00  43.382528 -78.230656
    4    1 2019-01-03 21:00:00  43.309133 -77.680414
    >>> print(ditras_tdf.parameters)
    {'model': {'class': <function Ditras.__init__ at 0x7f0cf0b7e158>, 'generate': {'start_date': Timestamp('2019-01-01 08:00:00'), 'end_date': Timestamp('2019-01-14 08:00:00'), 'gravity_singly': {}, 'n_agents': 3, 'relevance_column': 'population', 'random_state': None, 'show_progress': True}}}
    
    References
    ----------
    .. [PS2018] Pappalardo, L. & Simini, F. (2018) Data-driven generation of spatio-temporal routines in human mobility. Data Mining and Knowledge Discovery 32, 787-829, https://link.springer.com/article/10.1007/s10618-017-0548-4
    
    See Also
    --------
    DensityEPR, MarkovDiaryGenerator
    """

    def __init__(self, diary_generator, name='Ditras model', rho=0.3, gamma=0.21):

        super().__init__()
        self._diary_generator = diary_generator
        self._name = name
        self._rho = rho
        self._gamma = gamma

    def generate(self, start_date, end_date, spatial_tessellation, gravity_singly={}, n_agents=1,
                 starting_locations=None, od_matrix=None,
                 relevance_column=constants.RELEVANCE,
                 random_state=None, log_file=None, show_progress=False):
        """
        Start the simulation of a set of agents at time `start_date` till time `end_date`.
        
        Parameters
        ----------
        start_date : datetime
            the starting date of the simulation, in "YYY/mm/dd HH:MM:SS" format.

        end_date : datetime
            the ending date of the simulation, in "YYY/mm/dd HH:MM:SS" format.

        spatial_tessellation : geopandas GeoDataFrame
            the spatial tessellation, i.e., a division of the territory in locations. 
        
        gravity_singly : {} or Gravity, optional
            the gravity model (singly constrained or doubly constrained) to use when generating the probability to move between two locations. The default is "{}".
        
        n_agents : int, optional
            the number of agents to generate. The default is 1.

        starting_locations : list or None, optional
            a list of integers, each identifying the location from which to start the simulation of each agent. Note that, if `starting_locations` is not None, its length must be equal to the value of `n_agents`, i.e., you must specify one starting location per agent. The default is None.
        
        od_matrix : numpy array or None, optional
            the origin destination matrix to use for deciding the movements of the agent. If `None`, it is computed "on the fly" during the simulation. The default is None.
        
        relevance_column : str, optional
            the name of the column in `spatial_tessellation` to use as relevance variable. The default is "relevance".
        
        random_state : int or None, optional
            if int, it is the seed used by the random number generator; if None, the random number generator is the RandomState instance used by np.random and random.random. The default is None.
        
        log_file : str or None, optional
            the name of the file where to write a log of the execution of the model. The logfile will contain all decisions (returns or explorations) made by the model. The default is None.
            
        show_progress : boolean, optional
            if True, show a progress bar. The default is False.
        
        Returns
        -------
        TrajDataFrame
            the synthetic trajectories generated by the model
        """
        if starting_locations is not None and len(starting_locations) < n_agents:
            raise IndexError("The number of starting locations is smaller than the number of agents.")

        if gravity_singly == {}:
            self.gravity_singly = Gravity(gravity_type='singly constrained')

        # Save function arguments and values in a dictionary
        frame = inspect.currentframe()
        args, _, _, arg_values = inspect.getargvalues(frame)
        parameters = dict([])
        parameters['model'] = {'class': self.__class__.__init__,
                               'generate': {i: arg_values[i] for i in args[1:] if i not in ['spatial_tessellation',
                                                                                            'od_matrix', 'log_file',
                                                                                            'starting_locations']}}

        # if specified, fix the random seeds to guarantee reproducibility of simulation
        if random_state is not None:
            random.seed(random_state)
            np.random.seed(random_state)

        if log_file is not None:
            self._log_file = log_file
            logging.basicConfig(format='%(message)s', filename=log_file, filemode='w', level=logging.INFO)

        # initialization of trajectories
        self._trajectories_ = []

        # setting of spatial tessellation
        num_locs = len(spatial_tessellation)
        self.lats_lngs = spatial_tessellation.geometry.apply(utils.get_geom_centroid, args=[True]).values
        if relevance_column is None:
            self.relevances = np.ones(num_locs)
        else:
            self.relevances = spatial_tessellation[relevance_column].fillna(0).values

        # initialization of od matrix
        if od_matrix is None:
            self._od_matrix = lil_matrix((num_locs, num_locs))
            self._is_sparse = True
        else:
            self._od_matrix = od_matrix
            self._is_sparse = False

        # for each agent
        loop = range(1, n_agents + 1)
        if show_progress:
            loop = tqdm(range(1, n_agents + 1))

        for agent_id in loop:
            self._location2visits = defaultdict(int)
            if starting_locations is None:
                self._starting_loc = np.random.choice(np.fromiter(range(num_locs), dtype=int), size=1)[0]
            else:
                self._starting_loc = starting_locations.pop()

            self._epr_generate_one_agent(agent_id, start_date, end_date)

        tdf = self._get_trajdataframe(parameters)
        return tdf

        
    def _epr_generate_one_agent(self, agent_id, start_date, end_date):

        # infer the time_steps (in hours) from the start_date and the end_date
        delta_t = (end_date - start_date).total_seconds()
        n_hours = int((delta_t / 60.0) / 60.0)

        # generate a mobility diary for the agent
        diary_df = self._diary_generator.generate(n_hours, start_date)

        for i, row in diary_df.iterrows():
            if row.abstract_location == 0:  # the agent is at home
                self._trajectories_.append((agent_id, row.datetime, self._starting_loc))
                self._location2visits[self._starting_loc] += 1

            else:  # the agent is not at home
                next_location = self._choose_location()
                self._trajectories_.append((agent_id, row.datetime, next_location))
                self._location2visits[next_location] += 1



import numpy as np
from scipy import stats
import pandas as pd
from datetime import timedelta
from collections import defaultdict
from skmob.measures.individual import home_location
from ..utils import constants
import sys
from tqdm import tqdm
tqdm.pandas()
from skmob.utils.gislib import getDistanceByHaversine

def _random_location_entropy_individual(traj):
    """
    Compute the random location entropy of a single individual given their TrajDataFrame.

    Parameters
    ----------
    traj : TrajDataFrame
        the trajectory of the individual.
    
    Returns
    -------
    float
        the random location entropy of the individual.
    """
    n_distinct_users = len(traj.groupby(constants.UID))
    entropy = np.log2(n_distinct_users)
    return entropy

def random_location_entropy(traj, show_progress=True):
    """Random location entropy.
    
    Compute the random location entropy of the locations in a TrajDataFrame. The random location entropy of a location :math:`j` captures the degree of predictability of :math:`j` if each individual visits it with equal probability, and it is defined as:
    
    .. math::
        LE_{rand}(j) = log_2(N_j)
        
    where :math:`N_j` is the number of distinct individuals that visited location :math:`j`.
    
    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individuals.
    
    show_progress : boolean, optional
        if True, show a progress bar. The default is True.
    
    Returns
    -------
    pandas DataFrame
        the random location entropy of the locations.
        
    Example
    -------
    >>> import skmob
    >>> from skmob.measures.collective import random_location_entropy
    >>> url = "https://snap.stanford.edu/data/loc-brightkite_totalCheckins.txt.gz"
    >>> df = pd.read_csv(url, sep='\\t', header=0, nrows=100000, 
                 names=['user', 'check-in_time', 'latitude', 'longitude', 'location id'])
    >>> tdf = skmob.TrajDataFrame(df, latitude='latitude', longitude='longitude', datetime='check-in_time', user_id='user')
    >>> rle_df = random_location_entropy(tdf, show_progress=True).sort_values(by='random_location_entropy', ascending=False)
    >>> print(rle_df.head())
                 lat         lng  random_location_entropy
    10286  39.739154 -104.984703                 6.129283
    49      0.000000    0.000000                 5.643856
    5991   37.774929 -122.419415                 5.523562
    12504  39.878664 -104.682105                 5.491853
    5377   37.615223 -122.389979                 5.247928   
    
    See Also
    --------
    uncorrelated_location_entropy
    """
    # if 'uid' column in not present in the TrajDataFrame
    if constants.UID not in traj.columns:
        all_locations = traj[[constants.LATITUDE, constants.LONGITUDE]].drop_duplicates([constants.LATITUDE, constants.LONGITUDE])
        all_locations['random_location_entropy'] = 0.0
        return all_locations
    
    if show_progress:
        df = pd.DataFrame(traj.groupby([constants.LATITUDE, constants.LONGITUDE]).progress_apply(lambda x: _random_location_entropy_individual(x)))
    else:
        df = pd.DataFrame(traj.groupby([constants.LATITUDE, constants.LONGITUDE]).apply(lambda x: _random_location_entropy_individual(x)))
    column_name = sys._getframe().f_code.co_name
    return df.reset_index().rename(columns={0: column_name})

def _uncorrelated_location_entropy_individual(traj, normalize=True):
    n = len(traj)
    probs = [1.0 * len(group) / n for group in traj.groupby(by=constants.UID).groups.values()]
    entropy = stats.entropy(probs)
    if normalize:
        n_unique_users = len(traj[constants.UID].unique())
        if n_unique_users > 1:
            entropy /= np.log2(n_unique_users)
        else:  # to avoid NaN
            entropy = 0.0
    return entropy


def uncorrelated_location_entropy(traj, normalize=False, show_progress=True):
    """Temporal-uncorrelated entropy.
    
    Compute the temporal-uncorrelated location entropy of the locations in a TrajDataFrame. The temporal-uncorrelated location entropy :math:`LE_{unc}(j)` of a location :math:`j` is the historical probability that :math:`j` is visited by an individual :math:`u`. Formally, it is defined as [CML2011]_:
    
    .. math::
        LE_{unc}(j) = -\sum_{i=j}^{N_j} p_jlog_2(p_j)
        
    where :math:`N_j` is the number of distinct individuals that visited :math:`j` and :math:`p_j` is the historical probability that a visit to location :math:`j` is by individual :math:`u`. 
    
    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individuals.
    
    normalize : boolean, optional
        if True, normalize the location entropy by dividing by :math:`log2(N_j)`, where :math:`N_j` is the number of
        distinct individuals that visited location :math:`j`. The default is False.
    
    show_progress : boolean
        if True, show a progress bar. The default is True.
    
    Returns
    -------
    pandas DataFrame
        the temporal-uncorrelated location entropies of the locations.

    Examples
    --------
    >>> import skmob
    >>> from skmob.measures.collective import uncorrelated_location_entropy
    >>> url = "https://snap.stanford.edu/data/loc-brightkite_totalCheckins.txt.gz"
    >>> df = pd.read_csv(url, sep='\\t', header=0, nrows=100000, 
                 names=['user', 'check-in_time', 'latitude', 'longitude', 'location id'])
    >>> tdf = skmob.TrajDataFrame(df, latitude='latitude', longitude='longitude', datetime='check-in_time', user_id='user')
    >>> ule_df = uncorrelated_location_entropy(tdf, show_progress=True).sort_values(by='uncorrelated_location_entropy', ascending=False)
    >>> print(ule_df.head())
                 lat         lng  uncorrelated_location_entropy
    12504  39.878664 -104.682105                       3.415713
    5377   37.615223 -122.389979                       3.176950
    10286  39.739154 -104.984703                       3.118656
    12435  39.861656 -104.673177                       2.918413
    12361  39.848233 -104.675031                       2.899175
    dtype: float64
    
    References
    ----------
    .. [CML2011] Cho, E., Myers, S. A. & Leskovec, J. (2011) Friendship and mobility: user movement in location-based social networks. In Proceedings of the 17th ACM SIGKDD international conference on Knowledge discovery and data mining, 1082-1090, https://dl.acm.org/citation.cfm?id=2020579
        
    See Also
    --------
    random_location_entropy
    """
    # if 'uid' column in not present in the TrajDataFrame
    if constants.UID not in traj.columns:
        all_locations = traj[[constants.LATITUDE, constants.LONGITUDE]].drop_duplicates([constants.LATITUDE, constants.LONGITUDE])
        all_locations['uncorrelated_location_entropy'] = 0.0
        return all_locations
    
    if show_progress:
        df = pd.DataFrame(traj.groupby([constants.LATITUDE, constants.LONGITUDE]).progress_apply(lambda x: _uncorrelated_location_entropy_individual(x, normalize=normalize)))
    else:
        df = pd.DataFrame(traj.groupby([constants.LATITUDE, constants.LONGITUDE]).apply(lambda x: _uncorrelated_location_entropy_individual(x, normalize=normalize)))
    column_name = sys._getframe().f_code.co_name
    if normalize:
        column_name = 'norm_%s' % sys._getframe().f_code.co_name
    return df.reset_index().rename(columns={0: column_name})

def _square_displacement(traj, delta_t):
    """
    Compute the square displacement after time `delta_t` of a single individual given their TrajDataFrame.
    
    Parameters
    ----------
    traj : TrajDataFrame
        the trajectory of the individual.
        
    delta_t : datetime.timedelta
        the time from the reference position
        
    Returns
    -------
    float
        the square displacement of the individual.
    """
    r0 = traj.iloc[0]
    t = r0[constants.DATETIME] + delta_t
    rt = traj[traj[constants.DATETIME] <= t].iloc[-1]
    square_displacement = getDistanceByHaversine((r0.lat, r0.lng), (rt.lat, rt.lng)) ** 2
    return square_displacement


def mean_square_displacement(traj, days=0, hours=1, minutes=0, show_progress=True):
    """Mean Square Displacement.
    
    Compute the mean square displacement across the individuals in a TrajDataFrame. The mean squared displacement is a measure of the deviation of the position of an object with respect to a reference position over time [BHG2006]_ [SKWB2010]_. It is defined as:
    
    .. math::
        MSD = \\langle |r(t) - r(0)| \\rangle = \\frac{1}{N} \sum_{i = 1}^N |r^{(i)}(t) - r^{(i)}(0)|^2

    where :math:`N` is the number of individuals to be averaged, vector :math:`x^{(i)}(0)` is the reference position of the :math:`i`-th individual, and vector :math:`x^{(i)}(t)` is the position of the :math:`i`-th individual at time :math:`t` [FS2002]_.
    
    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individuals.
    
    days : int, optional
        the days since the starting time. The default is 0.
    
    hours : int, optional
        the hours since the days since the starting time. The default is 1.
    
    minutes : int, optional
        the minutes since the hours since the days since the starting time. The default is 0.

    show_progress : boolean, optional
        if True, show a progress bar. The default is True. 
    
    Returns
    -------
    float
        the mean square displacement.

    Warning
    -------
    The input TrajDataFrame must be sorted in ascending order by `datetime`.

    Examples
    --------
    >>> import skmob
    >>> url = "https://snap.stanford.edu/data/loc-brightkite_totalCheckins.txt.gz"
    >>> df = pd.read_csv(url, sep='\\t', header=0, nrows=100000, names=['user', 'check-in_time', 'latitude', 'longitude', 'location id'])
    >>> tdf = skmob.TrajDataFrame(df, latitude='latitude', longitude='longitude', datetime='check-in_time', user_id='user').sort_values(by='datetime')
    >>> msd = mean_square_displacement(tdf, days=0, hours=1, minutes=0)
    >>> print(msd)
    534672.3361996822

    References
    ----------
    .. [FS2002] Frenkel, D. & Smit, B. (2002) Understanding molecular simulation: From algorithms to applications. Academic Press, 196 (2nd Ed.), https://www.sciencedirect.com/book/9780122673511/understanding-molecular-simulation.
    .. [BHG2006] Brockmann, D., Hufnagel, L. & Geisel, T. (2006) The scaling laws of human travel. Nature 439, 462-465, https://www.nature.com/articles/nature04292
    .. [SKWB2010] Song, C., Koren, T., Wang, P. & Barabasi, A.L. (2010) Modelling the scaling properties of human mobility. Nature Physics 6, 818-823, https://www.nature.com/articles/nphys1760
    """
    delta_t = timedelta(days=days, hours=hours, minutes=minutes)
    
    # if 'uid' column in not present in the TrajDataFrame
    if constants.UID not in traj.columns:
        return _square_displacement(traj, delta_t)
    
    if show_progress:
        return traj.groupby(constants.UID).progress_apply(lambda x: _square_displacement(x, delta_t)).mean()
    else:
        return traj.groupby(constants.UID).apply(lambda x: _square_displacement(x, delta_t)).mean()


def visits_per_location(traj):
    """Visits per location.
    
    Compute the number of visits to each location in a TrajDataFrame [PF2018]_.
    
    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individuals.
    
    Returns
    -------
    pandas DataFrame
        the number of visits per location.

    Examples
    --------
    >>> import skmob
    >>> url = "https://snap.stanford.edu/data/loc-brightkite_totalCheckins.txt.gz"
    >>> df = pd.read_csv(url, sep='\\t', header=0, nrows=100000, names=['user', 'check-in_time', 'latitude', 'longitude', 'location id'])
    >>> tdf = skmob.TrajDataFrame(df, latitude='latitude', longitude='longitude', datetime='check-in_time', user_id='user').sort_values(by='datetime')
    >>> vl_df = visits_per_location(df)
    >>> print(vl_df.head())
             lat         lng  n_visits
    0  39.739154 -104.984703      3392
    1  37.580304 -122.343679      2248
    2  39.099275  -76.848306      1715
    3  39.762146 -104.982480      1442
    4  40.014986 -105.270546      1310

    References
    ----------
    .. [PF2018] Pappalardo, L. & Simini, F. (2018) Data-driven generation of spatio-temporal routines in human mobility. Data Mining and Knowledge Discovery 32, 787-829, https://link.springer.com/article/10.1007/s10618-017-0548-4
        
    See also
    --------
    homes_per_location
    """
    return traj.groupby([constants.LATITUDE,
                          constants.LONGITUDE]).count().sort_values(by=constants.DATETIME,
                                                                    ascending=False)[[constants.DATETIME]].reset_index().rename({constants.DATETIME: 'n_visits'}, axis=1)


def homes_per_location(traj, start_night='22:00', end_night='07:00'):
    """Homes per location.
    
    Compute the number of home locations in each location. The number of home locations in a location :math:`j` is computed as [PRS2016]_:
    
    .. math:: 
        N_{homes}(j) = |\{h_u | h_u = j, u \in U \}|

    where :math:`h_u` indicates the home location of an individual :math:`u` and :math:`U` is the set of individuals.
    
    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individuals.
    
    start_night : str, optional
        the starting time of the night (format HH:MM). The default is '22:00'.
        
    end_night : str, optional
        the ending time for the night (format HH:MM). The default is '07:00'.
    
    Returns
    -------
    pandas DataFrame
        the number of homes per location.

    Examples
    --------
    >>> import skmob
    >>> from skmob.measures.collective import homes_per_location
    >>> url = "https://snap.stanford.edu/data/loc-brightkite_totalCheckins.txt.gz"
    >>> df = pd.read_csv(url, sep='\\t', header=0, nrows=100000, names=['user', 'check-in_time', 'latitude', 'longitude', 'location id'])
    >>> tdf = skmob.TrajDataFrame(df, latitude='latitude', longitude='longitude', datetime='check-in_time', user_id='user').sort_values(by='datetime')
    >>> hl_df = homes_per_location(tdf).sort_values(by='n_homes', ascending=False)
    >>> print(hl_df.head())
             lat         lng  n_homes
    0  39.739154 -104.984703       15
    1  37.584103 -122.366083        6
    2  40.014986 -105.270546        5
    3  37.580304 -122.343679        5
    4  37.774929 -122.419415        4

    References
    ----------
    .. [PRS2016] Pappalardo, L., Rinzivillo, S. & Simini, F. (2016) Human Mobility Modelling: exploration and preferential return meet the gravity model. Procedia Computer Science 83, 934-939, http://dx.doi.org/10.1016/j.procs.2016.04.188
    """
    # if column 'uid' is not present in the TrajDataFrame
    uid_flag = False
    if constants.UID not in traj.columns:
        traj['uid'] = 0
        uid_flag = True
    df = home_location(traj,
                         start_night=start_night,
                         end_night=end_night).groupby([constants.LATITUDE,
                                                       constants.LONGITUDE]).count().sort_values(constants.UID,
                                                                                                 ascending=False).reset_index().rename(
        columns={constants.UID: 'n_homes'})
    if uid_flag:
        traj.drop('uid', axis=1, inplace=True)
    return df


def visits_per_time_unit(traj, time_unit='1h'):
    """Visits per time unit.
    
    Compute the number of data points per time unit in the TrajDataFrame [PRS2016]_.
    
    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individuals.
    
    time_unit : str, optional
        the time unit to use for grouping the time slots. The default '1h', which creates slots of 1 hour. For full specification of available time units, see http://pandas.pydata.org/pandas-docs/stable/timeseries.html#offset-aliases
    
    Returns
    -------
    pandas Series
        the number of visits per time unit.

    Examples
    --------
    >>> import skmob
    >>> from skmob.measures.collective import visits_per_time_unit
    >>> url = "https://snap.stanford.edu/data/loc-brightkite_totalCheckins.txt.gz"
    >>> df = pd.read_csv(url, sep='\\t', header=0, nrows=100000, names=['user', 'check-in_time', 'latitude', 'longitude', 'location id'])
    >>> tdf = skmob.TrajDataFrame(df, latitude='latitude', longitude='longitude', datetime='check-in_time', user_id='user').sort_values(by='datetime')
    >>> vtu_df = visits_per_time_unit(df)
    >>> print(vtu_df.head())
                               n_visits
    datetime                           
    2008-03-22 05:00:00+00:00         2
    2008-03-22 06:00:00+00:00         2
    2008-03-22 07:00:00+00:00         0
    2008-03-22 08:00:00+00:00         0
    2008-03-22 09:00:00+00:00         0
    """
    return pd.DataFrame(traj.set_index(pd.DatetimeIndex(traj[constants.DATETIME])).groupby(pd.Grouper(freq=time_unit)).count()[constants.DATETIME]).rename(columns={constants.DATETIME: 'n_visits'})


def origin_destination_matrix(traj, self_loops=False, show_progress=True):
    """Origin-destination matrix.
    
    Compute an origin-destination matrix from the trajectories of the individuals [CDLR2011]_.
    
    Parameters
    ----------
    traj : TrajDataFrame
        the trajectories of the individuals.
    
    show_progress : boolean, optional
        if True, show a progress bar. The default is True.
    
    self_loops : boolean, optional
        if True, include sel loops. The default is False.
    
    Returns
    -------
    pandas DataFrame
        the dataframe describing the origin destination matrix.
    
    Warning
    -------
    Note that the TrajDataFrame must sorted in ascending order by `datetime`.
    
    .. deprecated:: 1.0.1
          `origin_destination_matrix` will be removed in scikit-mobility 1.0.1, use `TrajDataFrame.to_flowdataframe`.
    
    Examples
    --------
    >>> import skmob
    >>> from skmob.measures.collective import origin_destination_matrix
    >>> url = "https://snap.stanford.edu/data/loc-brightkite_totalCheckins.txt.gz"
    >>> df = pd.read_csv(url, sep='\\t', header=0, nrows=100000, names=['user', 'check-in_time', 'latitude', 'longitude', 'location id'])
    >>> tdf = skmob.TrajDataFrame(df, latitude='latitude', longitude='longitude', datetime='check-in_time', user_id='user').sort_values(by='datetime')
    >>> od_df = origin_destination_matrix(tdf)
    >>> print(of_df.head())
       lat_origin  lng_origin   lat_dest    lng_dest  n_trips
    0   37.774929 -122.419415  37.600747 -122.382376        2
    1   37.774929 -122.419415  37.630490 -122.411084        8
    2   37.774929 -122.419415  37.584103 -122.366083       50
    3   37.774929 -122.419415  37.613825 -122.486919        2
    4   37.774929 -122.419415  37.654656 -122.407750        3

    References
    ----------
    .. [CDLR2011] Calabrese, F., Di Lorenzo, G., Liu, L. & Ratti, C. (2011) Estimating Origin-Destination Flows Using Mobile Phone Location Data. IEEE Pervasive Computing 10(4), 36-44, https://ieeexplore.ieee.org/document/5871578
    """
    loc2loc2weight = defaultdict(lambda: defaultdict(lambda: 0))

    def _update_od_matrix(traj):
        lats_lngs = traj[[constants.LATITUDE, constants.LONGITUDE]].values

        i = 1
        for lat, lng in lats_lngs[1:]:
            prev = tuple(lats_lngs[i - 1])
            current = (lat, lng)
            if prev != current:
                loc2loc2weight[prev][current] += 1
            elif self_loops:
                loc2loc2weight[prev][current] += 1
            else:
                pass
            i += 1
    
    # if 'uid' column in not present in the TrajDataFrame
    if constants.UID not in traj.columns:
        _update_od_matrix(traj.sort_values(by=constants.DATETIME))
    else:
        if show_progress:
            traj.sort_values(by=constants.DATETIME).groupby(constants.UID).progress_apply(lambda x: _update_od_matrix(x))
        else:
            traj.sort_values(by=constants.DATETIME).groupby(constants.UID).apply(lambda x: _update_od_matrix(x))
    rows = []
    for loc1, loc2weight in loc2loc2weight.items():
        for loc2, weight in loc2weight.items():
            rows.append([loc1[0], loc1[1], loc2[0], loc2[1], weight])
    return pd.DataFrame(rows, columns=[constants.LATITUDE + '_origin', constants.LONGITUDE + '_origin',
                                       constants.LATITUDE + '_dest', constants.LONGITUDE + '_dest', 'n_trips'])




from ..utils import gislib, utils, constants
from ..core.trajectorydataframe import *
import numpy as np
import inspect


def stops(tdf, stop_radius_factor=0.5, minutes_for_a_stop=20.0, spatial_radius_km=0.2, leaving_time=True, no_data_for_minutes=1e12, min_speed_kmh=None):
    """Stops detection.
    
    Detect the stops for each individual in a TrajDataFrame. A stop is detected when the individual spends at least `minutes_for_a_stop` minutes within a distance `stop_radius_factor * spatial_radius` km from a given trajectory point. The stop's coordinates are the median latitude and longitude values of the points found within the specified distance [RT2004]_ [Z2015]_.
    
    Parameters
    ----------
    tdf : TrajDataFrame
        the input trajectories of the individuals.

    stop_radius_factor : float, optional 
        the radius of the ball enclosing all trajectory points within the stop location. The default is `0.5`.

    minutes_for_a_stop : float, optional
        the minimum stop duration, in minutes. The default is `20.0`.

    spatial_radius_km : float or None, optional
        if `None`, use the spatial_radius specified in the TrajDataFrame properties (assigned by a `preprocessing.compression` function). The default is `0.2`.

    leaving_time : boolean, optional
        if `True`, a new column 'leaving_datetime' is added with the departure time from the stop location. The default is `True`.

    no_data_for_minutes : float, optional
        if the number of minutes between two consecutive points is larger than `no_data_for_minutes`,
        then this is interpreted as missing data and does not count as a stop. The default is `1e12`.

    min_speed_kmh : float or None, optional
        if not `None`, remove the points at the end of a stop if their speed is larger than `min_speed_kmh` km/h. The default is `None`.
    
    Returns
    -------
    TrajDataFrame
        a TrajDataFrame with the coordinates (latitude, longitude) of the stop locations.

    Examples
    --------
    >>> import skmob
    >>> import pandas as pd
    >>> from skmob.preprocessing import detection
    >>> # read the trajectory data (GeoLife)
    >>> url = 'https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/geolife_sample.txt.gz'
    >>> df = pd.read_csv(url, sep=',', compression='gzip')
    >>> tdf = skmob.TrajDataFrame(df, latitude='lat', longitude='lon', user_id='user', datetime='datetime')
    >>> print(tdf.head())
             lat         lng            datetime  uid
    0  39.984094  116.319236 2008-10-23 05:53:05    1
    1  39.984198  116.319322 2008-10-23 05:53:06    1
    2  39.984224  116.319402 2008-10-23 05:53:11    1
    3  39.984211  116.319389 2008-10-23 05:53:16    1
    4  39.984217  116.319422 2008-10-23 05:53:21    1
    >>> stdf = detection.stops(tdf, stop_radius_factor=0.5, minutes_for_a_stop=20.0, spatial_radius_km=0.2, leaving_time=True)
    >>> print(stdf.head())
             lat         lng            datetime  uid    leaving_datetime
    0  39.978030  116.327481 2008-10-23 06:01:37    1 2008-10-23 10:32:53
    1  40.013820  116.306532 2008-10-23 11:10:19    1 2008-10-23 23:45:27
    2  39.978419  116.326870 2008-10-24 00:21:52    1 2008-10-24 01:47:30
    3  39.981166  116.308475 2008-10-24 02:02:31    1 2008-10-24 02:30:29
    4  39.981431  116.309902 2008-10-24 02:30:29    1 2008-10-24 03:16:35
    >>> print(stdf.parameters)
    {'detect': {'function': 'stops', 'stop_radius_factor': 0.5, 'minutes_for_a_stop': 20.0, 'spatial_radius_km': 0.2, 'leaving_time': True, 'no_data_for_minutes': 1000000000000.0, 'min_speed_kmh': None}}
    >>> print('Points of the original trajectory:\\t%s'%len(tdf))
    >>> print('Points of stops:\\t\\t\\t%s'%len(stdf))
    Points of the original trajectory:	217653
    Points of stops:			391
    
    References
    ----------
    .. [RT2004] Ramaswamy, H. & Toyama, K. (2004) Project Lachesis: parsing and modeling location histories. In International Conference on Geographic Information Science, 106-124, http://kentarotoyama.com/papers/Hariharan_2004_Project_Lachesis.pdf
    .. [Z2015] Zheng, Y. (2015) Trajectory data mining: an overview. ACM Transactions on Intelligent Systems and Technology 6(3), https://dl.acm.org/citation.cfm?id=2743025
    """
    # Sort
    tdf = tdf.sort_by_uid_and_datetime()

    # Save function arguments and values in a dictionary
    frame = inspect.currentframe()
    args, _, _, arg_values = inspect.getargvalues(frame)
    arguments = dict([('function', stops.__name__)]+[(i, arg_values[i]) for i in args[1:]])

    groupby = []

    if utils.is_multi_user(tdf):
        groupby.append(constants.UID)
    if utils.is_multi_trajectory(tdf):
        groupby.append(constants.TID)

    # Use the spatial_radius in the tdf parameters, if present, otherwise use the default argument.
    try:
        spatial_radius_km = tdf.parameters[constants.COMPRESSION_PARAMS]['spatial_radius_km']
    except KeyError:
        pass
    stop_radius = spatial_radius_km * stop_radius_factor

    if len(groupby) > 0:
        # Apply simplify trajectory to each group of points
        stdf = tdf.groupby(groupby, group_keys=False, as_index=False).apply(_stops_trajectory, stop_radius=stop_radius,
                           minutes_for_a_stop=minutes_for_a_stop, leaving_time=leaving_time,
                           no_data_for_minutes=no_data_for_minutes, min_speed_kmh=min_speed_kmh).reset_index(drop=True)
    else:
        stdf = _stops_trajectory(tdf, stop_radius=stop_radius, minutes_for_a_stop=minutes_for_a_stop,
                            leaving_time=leaving_time, no_data_for_minutes=no_data_for_minutes,
                            min_speed_kmh=min_speed_kmh).reset_index(drop=True)

    # TODO: remove the following line when issue #71 (Preserve the TrajDataFrame index during preprocessing operations) is solved.
    stdf.reset_index(inplace=True, drop=True)

    stdf.parameters = tdf.parameters
    stdf.set_parameter(constants.DETECTION_PARAMS, arguments)
    return stdf


def _stops_trajectory(tdf, stop_radius, minutes_for_a_stop, leaving_time, no_data_for_minutes, min_speed_kmh):

    # From dataframe convert to numpy matrix
    lat_lng_dtime_other = list(utils.to_matrix(tdf))
    columns_order = list(tdf.columns)

    stops, leaving_times = _stops_array(lat_lng_dtime_other, stop_radius,
                                        minutes_for_a_stop, leaving_time, no_data_for_minutes, min_speed_kmh)

    #print(utils.get_columns(data))
    # stops = utils.to_dataframe(stops, utils.get_columns(data))
    stops = nparray_to_trajdataframe(stops, utils.get_columns(tdf), {})

    # Put back to the original order
    stops = stops[columns_order]

    if leaving_time:
        stops.loc[:, 'leaving_datetime'] = leaving_times

    return stops


def _stops_array(lat_lng_dtime_other, stop_radius, minutes_for_a_stop, leaving_time, no_data_for_minutes, min_speed_kmh):
    """
    Create a stop if the user spend at least `minutes_for_a_stop` minutes
    within a distance `stop_radius` from a given point.
    """
    # Define the distance function to use
    measure_distance = gislib.getDistance

    stops = []
    leaving_times = []

    lat_0, lon_0, t_0 = lat_lng_dtime_other[0][:3]
    sum_lat, sum_lon, sum_t = [lat_0], [lon_0], [t_0]
    speeds_kmh = []

    count = 1
    lendata = len(lat_lng_dtime_other) - 1

    for i in range(lendata):

        lat, lon, t = lat_lng_dtime_other[i+1][:3]

        if utils.diff_seconds(lat_lng_dtime_other[i][2], t) / 60. > no_data_for_minutes:
            # No data for more than `no_data_for_minutes` minutes: Not a stop
            count = 0
            lat_0, lon_0, t_0 = lat, lon, t
            sum_lat, sum_lon, sum_t = [], [], []
            speeds_kmh = []

        Dt = utils.diff_seconds(t_0, t) / 60.
        Dr = measure_distance([lat_0, lon_0], [lat, lon])
        try:
            speeds_kmh += [Dr / Dt * 60.]
        except ZeroDivisionError:
            speeds_kmh += [0.]

        if Dr > stop_radius:
            if Dt > minutes_for_a_stop:
                extra_cols = list(lat_lng_dtime_other[i][3:])

                # estimate the leaving time
                if min_speed_kmh is None:
                    estimated_final_t = t
                else:
                    j = 1
                    for j in range(1, len(speeds_kmh)):
                        if speeds_kmh[-j] < min_speed_kmh:
                            break
                    if j == 1:
                        estimated_final_t = t
                    else:
                        estimated_final_t = sum_t[-j]
                        sum_lat = sum_lat[:-j]
                        sum_lon = sum_lon[:-j]

                if len(sum_lat) > 0:
                    if leaving_time:
                        leaving_times.append(estimated_final_t)

                    stops += [[np.median(sum_lat), np.median(sum_lon), t_0] + extra_cols]

                count = 0
                lat_0, lon_0, t_0 = lat, lon, t
                sum_lat, sum_lon, sum_t = [], [], []
                speeds_kmh = []
            else:
                # Not a stop
                count = 0
                lat_0, lon_0, t_0 = lat, lon, t
                sum_lat, sum_lon, sum_t = [], [], []
                speeds_kmh = []

        count += 1
        sum_lat += [lat]
        sum_lon += [lon]
        sum_t += [t]

    return stops, leaving_times
